from collections.abc import Generator
from collections import defaultdict
from typing import Any, cast, Optional, Union
from collections.abc import Sequence
import warnings


from fandango.errors import FandangoValueError, FandangoParseError
from fandango.language.grammar import FuzzingMode, ParsingMode, closest_match
from fandango.language.grammar.has_settings import HasSettings
from fandango.language.grammar.literal_generator import LiteralGenerator
from fandango.language.grammar.node_visitors.disambiguator import Disambiguator
from fandango.language.grammar.node_visitors.node_visitor import NodeVisitor
from fandango.language.grammar.nodes.alternative import Alternative
from fandango.language.grammar.nodes.char_set import CharSet
from fandango.language.grammar.nodes.concatenation import Concatenation
from fandango.language.grammar.nodes.node import Node
from fandango.language.grammar.nodes.non_terminal import NonTerminalNode
import fandango.language.grammar.nodes as nodes
from fandango.language.grammar.nodes.repetition import (
    Option,
    Plus,
    Repetition,
    Star,
)
from fandango.language.grammar.nodes.terminal import TerminalNode
from fandango.language.grammar.parser.parser import Parser
from fandango.language.tree import DerivationTree, TreeTuple
from fandango.language.symbols import Symbol, NonTerminal
from fandango.language.tree_value import TreeValueType
from fandango.logger import LOGGER


class Grammar(NodeVisitor):
    """Represent a grammar."""

    def __init__(
        self,
        grammar_settings: Sequence[HasSettings],
        rules: Optional[dict[NonTerminal, Node]] = None,
        fuzzing_mode: Optional[FuzzingMode] = FuzzingMode.COMPLETE,
        local_variables: Optional[dict[str, Any]] = None,
        global_variables: Optional[dict[str, Any]] = None,
    ):
        self._grammar_settings = grammar_settings
        self.rules: dict[NonTerminal, Node] = rules or {}
        self.generators: dict[NonTerminal, LiteralGenerator] = {}
        self.fuzzing_mode = fuzzing_mode
        self._local_variables = local_variables or {}
        self._global_variables = global_variables or {}
        self._parser = Parser(self.rules)

    @property
    def grammar_settings(self) -> Sequence[HasSettings]:
        return self._grammar_settings

    @staticmethod
    def _topological_sort(graph: dict[NonTerminal, set[NonTerminal]]):
        indegree: dict[Any, int] = defaultdict(int)
        queue = []

        for node in graph:
            for neighbour in graph[node]:
                indegree[neighbour] += 1
        for node in graph:
            if indegree[node] == 0:
                queue.append(node)

        topological_order = []
        while queue:
            node = queue.pop(0)
            topological_order.append(node)

            for neighbour in graph[node]:
                indegree[neighbour] -= 1

                if indegree[neighbour] == 0:
                    queue.append(neighbour)

        if len(topological_order) != len(graph):
            print("Cycle exists")
        return topological_order[::-1]

    def is_use_generator(self, tree: DerivationTree):
        symbol = tree.symbol
        if not symbol.is_non_terminal:
            return False
        nt = cast(NonTerminal, symbol)
        if nt not in self.generators:
            return False
        if tree is None:
            path = set()
        else:
            path = set(map(lambda x: x.symbol, tree.get_path()))
        generator_dependencies = self.generator_dependencies(nt)
        intersection = path.intersection(set(generator_dependencies))
        return len(intersection) == 0

    def derive_sources(self, tree: DerivationTree) -> list[DerivationTree]:
        gen_symbol = tree.symbol
        if not gen_symbol.is_non_terminal:
            raise FandangoValueError(f"Tree {tree.symbol} is not a nonterminal")
        if tree.symbol not in self.generators:
            raise FandangoValueError(f"No generator found for tree {tree.symbol}")

        if not self.is_use_generator(tree):
            return []

        assert isinstance(gen_symbol, NonTerminal)
        dependent_generators: dict[NonTerminal, set[NonTerminal]] = {gen_symbol: set()}
        for key, val in self.generators[gen_symbol].nonterminals.items():
            if val.symbol not in self.rules:
                closest = closest_match(str(val), self.rules.keys())
                raise FandangoValueError(
                    f"Symbol {val.symbol.format_as_spec()} not defined in grammar. Did you mean {closest.format_as_spec()}?"
                )

            if val.symbol not in self.generators:
                raise FandangoValueError(
                    f"{val.symbol.format_as_spec()}: Missing converter from {gen_symbol.format_as_spec()} ({val.symbol.format_as_spec()} ::= ... := f({gen_symbol.format_as_spec()}))"
                )

            dependent_generators[val.symbol] = self.generator_dependencies(val.symbol)
        dependent_gens = self._topological_sort(dependent_generators)
        dependent_gens.remove(gen_symbol)

        args = [tree]
        for symbol in dependent_gens:
            generated_param = self.generate(symbol, args)
            generated_param.sources = []
            generated_param._parent = tree
            for child in generated_param.children:
                self.populate_sources(child)
            args.append(generated_param)
        args.pop(0)
        return args

    def derive_generator_output(self, tree: DerivationTree):
        generated = self.generate(tree.nonterminal, tree.sources)
        return generated.children

    def populate_sources(self, tree: DerivationTree):
        self._rec_remove_sources(tree)
        self._populate_sources(tree)

    def _populate_sources(self, tree: DerivationTree):
        if self.is_use_generator(tree):
            tree.sources = self.derive_sources(tree)
            for child in tree.children:
                child.set_all_read_only(True)
            return
        for child in tree.children:
            self._populate_sources(child)

    def _rec_remove_sources(self, tree: DerivationTree):
        tree.sources = []
        for child in tree.children:
            self._rec_remove_sources(child)

    def generate_string(
        self,
        symbol: str | NonTerminal = "<start>",
        sources: Optional[list[DerivationTree]] = None,
    ) -> tuple[list[DerivationTree], str | bytes | TreeTuple[str]]:
        if isinstance(symbol, str):
            symbol = NonTerminal(symbol)
        if self.generators[symbol] is None:
            raise ValueError(f"{symbol.format_as_spec()}: no generator")

        sources_: dict[Symbol, DerivationTree]
        if sources is None:
            sources_ = dict()
        else:
            sources_ = {tree.symbol: tree for tree in sources}
        generator = self.generators[symbol]

        local_variables = self._local_variables.copy()
        for id, nonterminal in generator.nonterminals.items():
            if nonterminal.symbol not in sources_:
                raise FandangoValueError(
                    f"{nonterminal.symbol}: missing generator parameter"
                )
            local_variables[id] = sources_[nonterminal.symbol]

        return list(sources_.values()), eval(
            generator.call, self._global_variables, local_variables
        )

    def generator_dependencies(self, symbol: str | NonTerminal = "<start>"):
        if isinstance(symbol, str):
            symbol = NonTerminal(symbol)
        if self.generators[symbol] is None:
            return set()
        return set(
            map(lambda x: x.symbol, self.generators[symbol].nonterminals.values())
        )

    def generate(
        self,
        symbol: str | NonTerminal = "<start>",
        sources: Optional[list[DerivationTree]] = None,
    ) -> DerivationTree:
        if isinstance(symbol, str):
            symbol = NonTerminal(symbol)
        sources, string = self.generate_string(symbol, sources)
        if not (isinstance(string, (str, bytes, int, tuple))):
            raise TypeError(
                f"Generator {self.generators[symbol]} must return string, bytes, int, or tuple (returned {string!r})"
            )

        if isinstance(string, tuple):
            warnings.warn(
                "Returning a tree in the shape of a tuple from a generator is deprecated, as it is parsed into a tree, then immediately stringified and parsed against the grammar. You should instead return a str/bytes directly"
            )
            string = str(DerivationTree.from_tree(string))
        tree = self.parse(string, symbol)
        if tree is None:
            raise FandangoParseError(
                f"Could not parse {string!r} (generated by {self.generators[symbol]}) into {symbol.format_as_spec()}"
            )
        tree.sources = [p.deepcopy(copy_parent=False) for p in sources]
        return tree

    def collapse(self, tree: Optional[DerivationTree]) -> Optional[DerivationTree]:
        return self._parser.collapse(tree)

    def fuzz(
        self,
        start: str | NonTerminal = "<start>",
        max_nodes: int = 50,
        prefix_node: Optional[DerivationTree] = None,
    ) -> DerivationTree:
        if isinstance(start, str):
            start = NonTerminal(start)
        if prefix_node is None:
            root = DerivationTree(start)
        else:
            root = prefix_node
        fuzzed_idx = len(root.children)
        NonTerminalNode(start, self._grammar_settings).fuzz(
            root, self, max_nodes=max_nodes
        )
        root = root.children[fuzzed_idx]
        root._parent = None
        return root

    def update(self, grammar: Union["Grammar", dict[NonTerminal, Node]], prime=True):
        generators: dict[NonTerminal, LiteralGenerator]
        local_variables: dict[str, Any]
        global_variables: dict[str, Any]
        if isinstance(grammar, Grammar):
            generators = grammar.generators
            local_variables = grammar._local_variables
            global_variables = grammar._global_variables
            rules = grammar.rules
            fuzzing_mode = grammar.fuzzing_mode
        else:
            rules = grammar
            generators = {}
            local_variables = {}
            global_variables = {}
            fuzzing_mode = FuzzingMode.COMPLETE

        self.rules.update(rules)
        self.fuzzing_mode = fuzzing_mode
        self.generators.update(generators)

        for symbol in rules.keys():
            # We're updating from a grammar with a rule, but no generator,
            # so we should remove the generator if it exists
            if symbol not in generators and symbol in self.generators:
                del self.generators[symbol]

        self._parser = Parser(self.rules)
        self._local_variables.update(local_variables)
        self._global_variables.update(global_variables)
        if prime:
            self.prime()

    def parse(
        self,
        word: str | bytes | int | DerivationTree,
        start: str | NonTerminal = "<start>",
        mode: ParsingMode = ParsingMode.COMPLETE,
        hookin_parent: Optional[DerivationTree] = None,
        include_controlflow: bool = False,
    ):
        return self._parser.parse(
            word,
            start,
            mode=mode,
            hookin_parent=hookin_parent,
            include_controlflow=include_controlflow,
        )

    def parse_forest(
        self,
        word: str | bytes | DerivationTree,
        start: str | NonTerminal = "<start>",
        mode: ParsingMode = ParsingMode.COMPLETE,
        include_controlflow: bool = False,
    ) -> Generator[Optional[DerivationTree], None, None]:
        return self._parser.parse_forest(
            word, start, mode=mode, include_controlflow=include_controlflow
        )

    def parse_multiple(
        self,
        word: str | bytes | DerivationTree,
        start: str | NonTerminal = "<start>",
        mode: ParsingMode = ParsingMode.COMPLETE,
        include_controlflow: bool = False,
    ):
        return self._parser.parse_multiple(
            word, start, mode=mode, include_controlflow=include_controlflow
        )

    def max_position(self):
        """Return the maximum position reached during last parsing."""
        return self._parser._iter_parser.max_position()

    def __contains__(self, item: str | NonTerminal):
        if not isinstance(item, NonTerminal):
            item = NonTerminal(item)
        return item in self.rules

    def __getitem__(self, item: str | NonTerminal):
        if not isinstance(item, NonTerminal):
            item = NonTerminal(item)
        return self.rules[item]

    def __setitem__(self, key: str | NonTerminal, value: Node):
        if not isinstance(key, NonTerminal):
            key = NonTerminal(key)
        self.rules[key] = value

    def __delitem__(self, key: str | NonTerminal):
        if not isinstance(key, NonTerminal):
            key = NonTerminal(key)
        del self.rules[key]

    def __iter__(self):
        return iter(self.rules)

    def __len__(self):
        return len(self.rules)

    def __repr__(self):
        return "\n".join(
            [
                f"{key.name()} ::= {value.format_as_spec()}{' := ' + str(self.generators[key]) if key in self.generators else ''}"
                for key, value in self.rules.items()
            ]
        )

    def msg_parties(self, *, include_recipients: bool = True) -> set:
        parties: set[str] = set()
        for rule in self.rules.values():
            parties |= rule.msg_parties(include_recipients=include_recipients)
        return parties

    def get_repr_for_rule(self, symbol: str | NonTerminal):
        if isinstance(symbol, str):
            symbol = NonTerminal(symbol)
        return (
            f"{symbol.format_as_spec()} ::= {self.rules[symbol].format_as_spec()}"
            f"{' := ' + str(self.generators[symbol]) if symbol in self.generators else ''}"
        )

    @staticmethod
    def dummy():
        return Grammar(grammar_settings=[], rules={})

    def set_generator(
        self, symbol: str | NonTerminal, param: str, searches_map: dict = {}
    ):
        if isinstance(symbol, str):
            symbol = NonTerminal(symbol)
        self.generators[symbol] = LiteralGenerator(
            call=param, nonterminals=searches_map
        )

    def remove_generator(self, symbol: str | NonTerminal):
        if isinstance(symbol, str):
            symbol = NonTerminal(symbol)
        self.generators.pop(symbol)

    def has_generator(self, symbol: str | NonTerminal):
        if isinstance(symbol, str):
            symbol = NonTerminal(symbol)
        return symbol in self.generators

    def get_generator(self, symbol: str | NonTerminal):
        if isinstance(symbol, str):
            symbol = NonTerminal(symbol)
        return self.generators.get(symbol, None)

    def update_parser(self):
        self._parser = Parser(self.rules)

    def compute_kpath_coverage(
        self, derivation_trees: list[DerivationTree], k: int
    ) -> float:
        """
        Computes the k-path coverage of the grammar given a set of derivation trees.
        Returns a score between 0 and 1 representing the fraction of k-paths covered.
        """
        # Generate all possible k-paths in the grammar
        all_k_paths = self._generate_all_k_paths(k)

        # Extract k-paths from the derivation trees
        covered_k_paths = set()
        for tree in derivation_trees:
            covered_k_paths.update(self._extract_k_paths_from_tree(tree, k))

        # Compute coverage score
        if not all_k_paths:
            return 1.0  # If there are no k-paths, coverage is 100%
        return len(covered_k_paths) / len(all_k_paths)

    def _generate_all_k_paths(self, k: int) -> set[tuple[Node, ...]]:
        """
        Computes the *k*-paths for this grammar, constructively. See: doi.org/10.1109/ASE.2019.00027

        :param k: The length of the paths.
        :return: All paths of length up to *k* within this grammar.
        """

        initial = set()
        initial_work: list[Node] = [
            NonTerminalNode(name, self._grammar_settings) for name in self.rules.keys()
        ]
        while initial_work:
            node = initial_work.pop(0)
            if node in initial:
                continue
            initial.add(node)
            initial_work.extend(node.descendents(self))

        work: list[set[tuple[Node, ...]]] = [set((x,) for x in initial)]

        for _ in range(1, k):
            next_work = set()
            for base in work[-1]:
                for descendent in base[-1].descendents(self):
                    next_work.add(base + (descendent,))
            work.append(next_work)

        # return set.union(*work)
        return work[-1]

    @staticmethod
    def _extract_k_paths_from_tree(
        tree: DerivationTree, k: int
    ) -> set[tuple[Symbol, ...]]:
        """
        Extracts all k-length paths (k-paths) from a derivation tree.
        """
        paths = set()

        def traverse(node: DerivationTree, current_path: tuple[Symbol, ...]):
            new_path = current_path + (node.symbol,)
            if len(new_path) == k:
                paths.add(new_path)
                # Do not traverse further to keep path length at k
                return
            for child in node.children:
                traverse(child, new_path)

        traverse(tree, ())
        return paths

    def prime(self):
        LOGGER.debug("Priming grammar")
        nodes: list[Node] = sum(
            [self.visit(self.rules[symbol]) for symbol in self.rules], []
        )
        while nodes:
            node = nodes.pop(0)
            if isinstance(node, TerminalNode):
                continue
            elif isinstance(node, NonTerminalNode):
                if node.symbol not in self.rules:
                    raise FandangoValueError(
                        f"Symbol {node.symbol.format_as_spec()} not found in grammar"
                    )
                if self.rules[node.symbol].distance_to_completion == float("inf"):
                    nodes.append(node)
                else:
                    node.distance_to_completion = (
                        self.rules[node.symbol].distance_to_completion + 1
                    )
            elif isinstance(node, Alternative):
                node.distance_to_completion = (
                    min([n.distance_to_completion for n in node.alternatives]) + 1
                )
                if node.distance_to_completion == float("inf"):
                    nodes.append(node)
            elif isinstance(node, Concatenation):
                if any([n.distance_to_completion == float("inf") for n in node.nodes]):
                    nodes.append(node)
                else:
                    node.distance_to_completion = (
                        sum([n.distance_to_completion for n in node.nodes]) + 1
                    )
            elif isinstance(node, Repetition):
                if node.node.distance_to_completion == float("inf"):
                    nodes.append(node)
                else:
                    node.distance_to_completion = (
                        node.node.distance_to_completion * node.min + 1
                    )
            else:
                raise FandangoValueError(f"Unknown node type {node.node_type}")

    def slice_parties(self, parties: list[str]) -> None:
        """
        Returns a new grammar that only contains the rules that are relevant to the given parties.
        """
        for expansion in self.rules.values():
            expansion.slice_parties(parties)
        self.fuzzing_mode = FuzzingMode.COMPLETE

    def default_result(self):
        return []

    def aggregate_results(self, aggregate, result):
        aggregate.extend(result)
        return aggregate

    def visitAlternative(self, node: Alternative):
        return self.visitChildren(node) + [node]

    def visitConcatenation(self, node: Concatenation):
        return self.visitChildren(node) + [node]

    def visitRepetition(self, node: Repetition):
        return self.visit(node.node) + [node]

    def visitStar(self, node: Star):
        return self.visit(node.node) + [node]

    def visitPlus(self, node: Plus):
        return self.visit(node.node) + [node]

    def visitOption(self, node: Option):
        return self.visit(node.node) + [node]

    def visitNonTerminalNode(self, node: NonTerminalNode):
        return [node]

    def visitTerminalNode(self, node: TerminalNode):
        return []

    def visitCharSet(self, node: CharSet):
        return []

    def compute_k_paths(self, k: int) -> set[tuple[Node, ...]]:
        """
        Computes all possible k-paths in the grammar.

        :param k: The length of the paths.
        :return: A set of tuples, each tuple representing a k-path as a sequence of symbols.
        """
        return self._generate_all_k_paths(k)

    def traverse_derivation(
        self,
        tree: DerivationTree,
        disambiguator: Optional[Disambiguator] = None,
        paths: Optional[set[tuple[Node, ...]]] = None,
        cur_path: Optional[tuple[Node, ...]] = None,
    ) -> set[tuple[Node, ...]]:
        if disambiguator is None:
            disambiguator = Disambiguator(self, self._grammar_settings)
        if paths is None:
            paths = set()
        if tree.symbol.is_terminal:
            if cur_path is None:
                cur_path = (TerminalNode(tree.terminal, self._grammar_settings),)
            paths.add(cur_path)
        elif isinstance(tree.symbol, NonTerminal):
            if cur_path is None:
                cur_path = (NonTerminalNode(tree.nonterminal, self._grammar_settings),)
            assert tree.symbol == cast(NonTerminalNode, cur_path[-1]).symbol
            disambiguation = disambiguator.visit(self.rules[tree.nonterminal])
            for tree, path in zip(
                tree.children, disambiguation[tuple(c.symbol for c in tree.children)]
            ):
                self.traverse_derivation(tree, disambiguator, paths, cur_path + path)
        else:
            raise FandangoValueError(
                f"Unknown symbol type: {type(tree.symbol)}: {tree.symbol}"
            )
        return paths

    def compute_grammar_coverage(
        self, derivation_trees: list[DerivationTree], k: int
    ) -> tuple[float, int, int]:
        """
        Compute the coverage of k-paths in the grammar based on the given derivation trees.

        :param derivation_trees: A list of derivation trees (solutions produced by FANDANGO).
        :param k: The length of the paths (k).
        :return: A float between 0 and 1 representing the coverage.
        """

        # Compute all possible k-paths in the grammar
        all_k_paths = self.compute_k_paths(k)

        disambiguator = Disambiguator(self, self._grammar_settings)

        # Extract k-paths from the derivation trees
        covered_k_paths = set()
        for tree in derivation_trees:
            for path in self.traverse_derivation(tree, disambiguator):
                # for length in range(1, k + 1):
                for window in range(len(path) - k + 1):
                    covered_k_paths.add(path[window : window + k])

        # Compute coverage
        if not all_k_paths:
            raise FandangoValueError("No k-paths found in the grammar")

        return (
            len(covered_k_paths) / len(all_k_paths),
            len(covered_k_paths),
            len(all_k_paths),
        )

    def get_spec_env(self):
        return self._global_variables, self._local_variables

    def contains_type(self, tp: TreeValueType, *, start="<start>") -> bool:
        """
        Return true if the grammar can produce an element of type `tp` (say, `int` or `bytes`).
        * `start`: a start symbol other than `<start>`.
        """
        if isinstance(tp, TreeValueType):
            tvt = tp
        else:
            if isinstance(tp, str):
                tvt = TreeValueType.STRING
            elif isinstance(tp, bytes):
                tvt = TreeValueType.BYTES
            elif isinstance(tp, int):
                tvt = TreeValueType.TRAILING_BITS_ONLY
            else:
                raise FandangoValueError(f"Invalid type: {type(tp)}")

        if isinstance(start, str):
            start = NonTerminal(start)

        if start not in self.rules:
            raise FandangoValueError(f"Start symbol {start} not defined in grammar")

        # We start on the right hand side of the start symbol
        start_node = self.rules[start]
        seen = set()

        def node_matches(node):
            if node in seen:
                return False
            seen.add(node)

            if isinstance(node, TerminalNode):
                if node.symbol.is_type(tvt):
                    return True
            if any(node_matches(child) for child in node.children()):
                return True
            if isinstance(node, NonTerminalNode):
                return node_matches(self.rules[node.symbol])
            return False

        return node_matches(start_node)

    def contains_bits(self, *, start="<start>") -> bool:
        """
        Return true iff the grammar can produce a bit element (0 or 1).
        * `start`: a start symbol other than `<start>`.
        """
        return self.contains_type(TreeValueType.TRAILING_BITS_ONLY, start=start)

    def contains_bytes(self, *, start="<start>") -> bool:
        """
        Return true iff the grammar can produce a bytes element.
        * `start`: a start symbol other than `<start>`.
        """
        return self.contains_type(TreeValueType.BYTES, start=start)

    def set_max_repetition(self, max_rep: int):
        nodes.MAX_REPETITIONS = max_rep

    def get_max_repetition(self):
        return nodes.MAX_REPETITIONS



from abc import ABC, abstractmethod
from typing import Any, Optional
from fandango.language.tree import DerivationTree
from fandango.constraints.fitness import ConstraintFitness, GeneticBase
from fandango.language.search import NonTerminalSearch
from fandango.language.symbols.non_terminal import NonTerminal


class Constraint(GeneticBase, ABC):
    """
    Abstract class to represent a constraint that can be used for fitness evaluation.
    """

    def __init__(
        self,
        searches: Optional[dict[str, NonTerminalSearch]] = None,
        local_variables: Optional[dict[str, Any]] = None,
        global_variables: Optional[dict[str, Any]] = None,
    ):
        """
        Initializes the constraint with the given searches, local variables, and global variables.
        :param Optional[dict[str, NonTerminalSearch]] searches: The searches to use.
        :param Optional[dict[str, Any]] local_variables: The local variables to use.
        :param Optional[dict[str, Any]] global_variables: The global variables to use.
        """
        super().__init__(searches, local_variables, global_variables)
        self.cache: dict[int, ConstraintFitness] = dict()

    @abstractmethod
    def fitness(
        self,
        tree: DerivationTree,
        scope: Optional[dict[NonTerminal, DerivationTree]] = None,
        population: Optional[list[DerivationTree]] = None,
        local_variables: Optional[dict[str, Any]] = None,
    ) -> ConstraintFitness:
        """
        Abstract method to calculate the fitness of the tree.
        """
        raise NotImplementedError("Fitness function not implemented")

    @staticmethod
    def is_debug_statement(expression: str) -> bool:
        """
        Determines if the expression is a print statement.
        """
        return expression.startswith("print(")

    @abstractmethod
    def accept(self, visitor):
        """
        Accepts a visitor to traverse the constraint structure.
        """
        pass

    def get_symbols(self):
        """
        Get the placeholders of the constraint.
        """
        return self.searches.values()

    @staticmethod
    def eval(expression: str, global_variables, local_variables):
        """
        Evaluate the tree in the context of local and global variables.
        """
        # LOGGER.debug(f"Evaluating {expression}")
        # for name, value in local_variables.items():
        #     if isinstance(value, DerivationTree):
        #         value = value.value()
        #     LOGGER.debug(f"    {name} = {value!r}")

        result = eval(expression, global_variables, local_variables)

        # res = result
        # if isinstance(res, DerivationTree):
        #     res = res.value()
        # LOGGER.debug(f"Result = {res!r}")

        return result

    @abstractmethod
    def format_as_spec(self) -> str:
        """
        Format the constraint as a string that can be used in a spec file.
        """

    def __repr__(self):
        raise NotImplementedError(
            "Repr not implemented, use method specific to your usecase"
        )



#!/usr/bin/env pytest

import glob
import unittest

import pytest

from .utils import RESOURCES_ROOT, DOCS_ROOT, run_command


files = glob.glob(str(RESOURCES_ROOT / "*.fan")) + glob.glob(str(DOCS_ROOT / "*.fan"))


@pytest.mark.parametrize("fan_file", files)
def test_file(fan_file):
    """Test the C++ and python .fan parsers for `fan_file`."""

    command = ["fandango", "-v", "--parser=python", "convert", fan_file]
    python_out, err, return_code = run_command(command)
    assert return_code == 0, err
    assert err == ""

    command = ["fandango", "--parser=cpp", "convert", fan_file]
    cpp_out, err, return_code = run_command(command)
    assert return_code == 0, err
    assert err == ""

    assert (
        python_out == cpp_out
    ), f"{fan_file} produced different outputs for Python and C++ parsers:\n\nPython output:\n{python_out}\n\nC++ output:\n{cpp_out}"


if __name__ == "__main__":
    unittest.main()



from typing import TYPE_CHECKING
from collections.abc import Iterable, Iterator, Sequence
from fandango.language.grammar.has_settings import HasSettings
from fandango.language.grammar.nodes.node import Node, NodeType
from fandango.language.tree import DerivationTree

if TYPE_CHECKING:
    import fandango


class Concatenation(Node):
    def __init__(
        self,
        nodes: Iterable[Node],
        grammar_settings: Sequence[HasSettings],
        id: str = "",
    ):
        self.id = id
        self.nodes = list(nodes)
        super().__init__(NodeType.CONCATENATION, grammar_settings)

    def fuzz(
        self,
        parent: DerivationTree,
        grammar: "fandango.language.grammar.grammar.Grammar",
        max_nodes: int = 100,
        in_message: bool = False,
    ):
        prev_parent_size = parent.size()
        for node in self.nodes:
            if node.distance_to_completion >= max_nodes:
                node.fuzz(parent, grammar, 0, in_message)
            else:
                reserved_distance = self.distance_to_completion
                for dist_node in self.nodes:
                    reserved_distance -= dist_node.distance_to_completion
                    if dist_node == node:
                        break
                node.fuzz(
                    parent, grammar, int(max_nodes - reserved_distance), in_message
                )
            max_nodes -= parent.size() - prev_parent_size
            prev_parent_size = parent.size()

    def accept(
        self,
        visitor: "fandango.language.grammar.node_visitors.node_visitor.NodeVisitor",
    ):
        return visitor.visitConcatenation(self)

    def children(self):
        return self.nodes

    def slice_parties(self, parties: list[str]) -> None:
        self.nodes = [node for node in self.nodes if node.in_parties(parties)]
        super().slice_parties(parties)

    def __getitem__(self, item):
        return self.nodes.__getitem__(item)

    def __len__(self):
        return len(self.nodes)

    def format_as_spec(self) -> str:
        return " ".join(map(lambda x: x.format_as_spec(), self.nodes))

    def descendents(
        self, grammar: "fandango.language.grammar.grammar.Grammar"
    ) -> Iterator["Node"]:
        yield from self.nodes



import abc
import itertools
from typing import TYPE_CHECKING, Any, Optional

from fandango.language.search import Container, NonTerminalSearch
from fandango.language.symbols.non_terminal import NonTerminal
from fandango.language.tree import DerivationTree

if TYPE_CHECKING:
    from fandango.constraints.fitness import FailingTree, Fitness


class GeneticBase(abc.ABC):
    """
    Abstract class to represent a genetic base.
    """

    def __init__(
        self,
        searches: Optional[dict[str, NonTerminalSearch]] = None,
        local_variables: Optional[dict[str, Any]] = None,
        global_variables: Optional[dict[str, Any]] = None,
    ):
        """
        Initialize the GeneticBase with the given searches, local variables, and global variables.
        :param Optional[dict[str, NonTerminalSearch]] searches: The dictionary of searches.
        :param Optional[dict[str, Any]] local_variables: The dictionary of local variables.
        :param Optional[dict[str, Any]] global_variables: The dictionary of global variables.
        """
        self.searches = searches or dict()
        self.local_variables = local_variables or dict()
        self.global_variables = global_variables or dict()

    def get_access_points(self):
        """
        Get the access points of the genetic base, i.e., the non-terminal that are considered in this genetic base.
        :return list[NonTerminal]: The list of access points.
        """
        return sum(
            [search.get_access_points() for search in self.searches.values()], []
        )

    @abc.abstractmethod
    def fitness(
        self,
        tree: DerivationTree,
        scope: Optional[dict[NonTerminal, DerivationTree]] = None,
        population: Optional[list[DerivationTree]] = None,
        local_variables: Optional[dict[str, Any]] = None,
    ) -> "Fitness":
        """
        Abstract method to calculate the fitness of the tree.
        :param DerivationTree tree: The tree to calculate the fitness.
        :param Optional[dict[NonTerminal, DerivationTree]] scope: The scope of non-terminals matching to trees.
        :param Optional[list[DerivationTree]] population: The population of trees to calculate the fitness.
        :param Optional[dict[str, Any]] local_variables: The local variables to use in the fitness calculation.
        :return Fitness: The fitness of the tree.
        """
        raise NotImplementedError("Fitness function not implemented")

    @staticmethod
    def get_hash(
        tree: DerivationTree,
        scope: Optional[dict[NonTerminal, DerivationTree]] = None,
        population: Optional[list[DerivationTree]] = None,
        local_variables: Optional[dict[str, Any]] = None,
    ):
        return hash(
            (
                tree,
                tuple((scope or {}).items()),
                tuple(population or []),
                tuple((local_variables or {}).items()),
            )
        )

    def combinations(
        self,
        tree: DerivationTree,
        scope: Optional[dict[NonTerminal, DerivationTree]] = None,
        population: Optional[list[DerivationTree]] = None,
    ):
        """
        Get all possible combinations of trees that satisfy the searches.
        :param DerivationTree tree: The tree to calculate the fitness.
        :param Optional[dict[NonTerminal, DerivationTree]] scope: The scope of non-terminals matching to trees.
        :return list[list[tuple[str, DerivationTree]]]: The list of combinations of trees that fill all non-terminals
        :param Optional[list[DerivationTree]] population: The population of trees to calculate the fitness.
        in the genetic base.
        """
        nodes: list[list[tuple[str, Container]]] = []
        for name, search in self.searches.items():
            nodes.append(
                [
                    (name, container)
                    for container in search.find(
                        tree, scope=scope, population=population
                    )
                ]
            )
        return itertools.product(*nodes)

    def check(
        self,
        tree: DerivationTree,
        scope: Optional[dict[NonTerminal, DerivationTree]] = None,
        population: Optional[list[DerivationTree]] = None,
        local_variables: Optional[dict[str, Any]] = None,
    ) -> bool:
        """
        Check if the tree satisfies the genetic base.
        :param DerivationTree tree: The tree to check.
        :param Optional[dict[NonTerminal, DerivationTree]] scope: The scope of non-terminals matching to trees.
        :param Optional[list[DerivationTree]] population: The population of trees to calculate the fitness.
        :param Optional[dict[str, Any]] local_variables: The local variables to use in the fitness calculation.
        :return bool: True if the tree satisfies the genetic base, False otherwise.
        """
        return self.fitness(tree, scope, population, local_variables).success

    def get_failing_nodes(
        self,
        tree: DerivationTree,
        scope: Optional[dict[NonTerminal, DerivationTree]] = None,
        population: Optional[list[DerivationTree]] = None,
        local_variables: Optional[dict[str, Any]] = None,
    ) -> list["FailingTree"]:
        """
        Get the failing nodes of the tree.
        :param DerivationTree tree: The tree to check.
        :param Optional[dict[NonTerminal, DerivationTree]] scope: The scope of non-terminals matching to trees.
        :param Optional[list[DerivationTree]] population: The population of trees to calculate the fitness.
        :param Optional[dict[str, Any]] local_variables: The local variables to use in the fitness calculation.
        :return list[FailingTree]: The list of failing trees
        """
        return self.fitness(tree, scope, population, local_variables).failing_trees

    def __str__(self) -> str:
        raise RuntimeError("Not implemented, use method specific to your usecase")

    def __repr__(self) -> str:
        raise RuntimeError("Not implemented, use method specific to your usecase")

    @abc.abstractmethod
    def format_as_spec(self) -> str:
        """
        Format as a string that can be used in a spec file.
        """



#!/usr/bin/env python3

import enum
import logging
import re
import select
import socket
import subprocess
import sys
import threading
import time
from abc import ABC
from typing import Optional

from fandango.errors import FandangoError, FandangoValueError
from fandango.language.tree import DerivationTree
from fandango.logger import LOGGER


class Protocol(enum.Enum):
    TCP = "TCP"
    UDP = "UDP"


class EndpointType(enum.Enum):
    OPEN = "Open"
    CONNECT = "Connect"


class IpType(enum.Enum):
    IPV4 = "IPv4"
    IPV6 = "IPv6"


class Ownership(enum.Enum):
    FANDANGO_PARTY = "FandangoParty"
    EXTERNAL_PARTY = "ExternalParty"


RE_PARTY = re.compile(
    r"""
((?P<name>[a-zA-Z0-9_]+)=)?            # Optional party name followed by =
((?P<protocol>([tT][Cc][pP]|[uU][dD][pP])):)?  # Optional protocol prefixed by :
(//)?                                  # Optional // separator
((?P<host>([^:]+|\[(?P<ipv6>.*)\])):)? # hostname(IPv6 in [...])
(?P<port>[0-9]+)                       # Port
""",
    re.VERBOSE,
)


def split_party_spec(
    spec: str,
) -> tuple[Optional[str], Optional[str], Optional[str], int]:
    """
    Splits a party specification into the party name and the party definition.
    :param spec: The party specification to split.
    :return: A tuple containing
        - The party name (str) or None if not specified
        - The party protocol (str) or None if not specified
        - The party host (str) or None if not specified
        - The party port (int)
    """
    match = RE_PARTY.match(spec)
    if not match:
        raise FandangoValueError(f"Invalid party specification: {spec}")
    name = match.group("name")
    host = match.group("ipv6") or match.group("host")
    port = int(match.group("port"))
    protocol = match.group("protocol")
    if protocol is not None:
        protocol = protocol.upper()
    return name, protocol, host, port


class FandangoParty(ABC):
    """Base class for all parties in Fandango."""

    def __init__(self, *, ownership: Ownership, party_name: Optional[str] = None):
        """Constructor.
        :param ownership: Ownership of the party, either Ownership.FUZZER or Ownership.EXTERNAL. FUZZER means the party is controlled by Fandango, while EXTERNAL means it is an external party.
        """
        if party_name is None:
            self.party_name = type(self).__name__
        else:
            self.party_name = party_name
        self._ownership = ownership
        FandangoIO.instance().parties[self.party_name] = self

    @property
    def ownership(self) -> Ownership:
        """
        :return: ownership of the party
        """
        return self._ownership

    def is_fuzzer_controlled(self) -> bool:
        """
        Returns True if this party is owned by Fandango, False if it is an external party.
        """
        return self.ownership == Ownership.FANDANGO_PARTY

    def on_send(self, message: DerivationTree, recipient: Optional[str]) -> None:
        """
        Called when fandango wants to send a message as this party.
        :param message: The message to send.
        :param recipient: The recipient of the message. Only present if the grammar specifies a recipient.
        """
        print(f"({self.party_name}): {message}")

    def receive_msg(self, sender: Optional[str], message: str | bytes) -> None:
        """
        Called when a message has been received by this party.
        :param sender: The sender of the message.
        :param message: The sender of the message.
        """
        if sender is None:
            parties = list(
                map(
                    lambda x: x.party_name,
                    filter(
                        lambda party: not party.is_fuzzer_controlled(),
                        FandangoIO.instance().parties.values(),
                    ),
                )
            )
            if len(parties) == 1:
                sender = parties[0]
            else:
                raise FandangoValueError(
                    f"Could not determine sender of message received by {self.party_name}. Please explicitly provide the sender to the receive_msg method."
                )
        FandangoIO.instance().add_receive(sender, self.party_name, message)


class ProtocolDecorator(ABC):
    def __init__(
        self,
        *,
        endpoint_type: EndpointType = EndpointType.CONNECT,
        ip_type: IpType = IpType.IPV4,
        ip: Optional[str],
        port: Optional[int],
        party_instance: FandangoParty,
    ):
        self.endpoint_type = endpoint_type
        self.ip = ip
        self.port = port
        self.ip_type = ip_type
        self._party_instance = party_instance

    def on_send(self, message: DerivationTree, recipient: Optional[str]):
        raise NotImplementedError("Please Implement this method")

    def start(self):
        raise NotImplementedError("Please Implement this method")

    def stop(self):
        raise NotImplementedError("Please Implement this method")

    @property
    def protocol_type(self) -> Protocol:
        raise NotImplementedError("Please Implement this method")


class UdpTcpProtocolDecorator(ProtocolDecorator):
    BUFFER_SIZE = 1024  # Size of the buffer for receiving data

    def __init__(
        self,
        *,
        endpoint_type: EndpointType = EndpointType.CONNECT,
        protocol_type: Protocol,
        ip_type: IpType = IpType.IPV4,
        ip: Optional[str] = None,
        port: Optional[int] = None,
        party_instance: Optional[FandangoParty] = None,
    ):
        if party_instance is None:
            raise FandangoValueError("party_instance must not be None")
        super().__init__(
            endpoint_type=endpoint_type,
            ip_type=ip_type,
            ip=ip,
            port=port,
            party_instance=party_instance,
        )
        self._running = False
        assert protocol_type == Protocol.TCP or protocol_type == Protocol.UDP
        self._protocol_type = protocol_type
        self._sock: Optional[socket.socket] = None
        self._connection: Optional[socket.socket] = None
        self._send_thread: Optional[threading.Thread] = None
        self.current_remote_addr = None
        self._lock = threading.Lock()

    @property
    def protocol_type(self) -> Protocol:
        """Returns the protocol type of this socket."""
        return self._protocol_type

    def start(self):
        """Starts the socket party according to the given configuration. If the party is already
        running or ownership is not set to Ownership.FUZZER, it does nothing."""
        if self._running:
            return
        if not self._party_instance.is_fuzzer_controlled():
            return
        self.stop()
        self._create_socket()
        self._connect()

    def _create_socket(self):
        protocol = (
            socket.SOCK_STREAM
            if self._protocol_type == Protocol.TCP
            else socket.SOCK_DGRAM
        )
        ip_type = socket.AF_INET if self.ip_type == IpType.IPV4 else socket.AF_INET6
        self._sock = socket.socket(ip_type, protocol)
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    def _connect(self):
        if self.endpoint_type == EndpointType.OPEN:
            assert self._sock is not None
            self._sock.bind((self.ip, self.port))
            if self.protocol_type == Protocol.TCP:
                self._sock.listen(1)
        self._running = True
        self._send_thread = threading.Thread(target=self._listen, daemon=True)
        self._send_thread.daemon = True
        self._send_thread.start()

    def stop(self):
        """Stops the current socket."""
        self._running = False
        if self._send_thread is not None:
            self._send_thread.join()
            self._send_thread = None
        if self._connection is not None:
            try:
                self._connection.shutdown(socket.SHUT_RDWR)
            except OSError:
                pass
            try:
                self._connection.close()
            except OSError:
                pass
            self._connection = None
        if self._sock is not None:
            try:
                self._sock.shutdown(socket.SHUT_RDWR)
            except OSError:
                pass
            try:
                self._sock.close()
            except OSError:
                pass
            self._sock = None

    def _wait_accept(self):
        with self._lock:
            if self._connection is None:
                if self.protocol_type == Protocol.TCP:
                    if self.endpoint_type == EndpointType.OPEN:
                        assert self._sock is not None
                        while self._running:
                            rlist, _, _ = select.select([self._sock], [], [], 0.1)
                            if rlist:
                                self._connection, _ = self._sock.accept()
                                break
                    else:
                        assert self._sock is not None
                        self._sock.setblocking(False)
                        try:
                            self._sock.connect((self.ip, self.port))
                        except BlockingIOError:
                            pass
                        while self._running:
                            _, wlist, _ = select.select([], [self._sock], [], 0.1)
                            if wlist:
                                self._connection = self._sock
                                break
                        self._sock.setblocking(True)
                else:
                    # For UDP, we do not need to accept a connection
                    assert self._sock is not None
                    self._connection = self._sock

    def _listen(self):
        self._wait_accept()
        if not self._running:
            return

        while self._running:
            try:
                assert self._connection is not None
                rlist, _, _ = select.select([self._connection], [], [], 0.1)
                if rlist and self._running:
                    if self.protocol_type == Protocol.TCP:
                        data = self._connection.recv(self.BUFFER_SIZE)
                    else:
                        data, addr = self._connection.recvfrom(self.BUFFER_SIZE)
                        self.current_remote_addr = addr
                    if len(data) == 0:
                        continue  # Keep waiting if connection is open but no data
                    self._party_instance.receive_msg(None, data)
            except Exception:
                self._running = False
                break

    def on_send(self, message: DerivationTree, recipient: Optional[str]):
        """Called when Fandango wants to send a message as this party.
        :param message: The message to send.
        :param recipient: The recipient of the message. Only present if the grammar specifies a recipient.
        :raises FandangoError: If the socket is not running.
        """
        if not self._running:
            raise FandangoError("Socket not running. Invoke start() first.")
        self._wait_accept()

        assert self._connection is not None
        send_data = message.to_bytes(encoding="utf-8")
        if self.protocol_type == Protocol.TCP:
            self._connection.sendall(send_data)
        else:
            if self.endpoint_type == EndpointType.OPEN:
                if self.current_remote_addr is None:
                    raise FandangoValueError(
                        "Client received no data yet. No address to send to."
                    )
                self._connection.sendto(send_data, self.current_remote_addr)
            else:
                self._connection.sendto(send_data, (self.ip, self.port))


class ConnectParty(FandangoParty):
    DEFAULT_IP = "127.0.0.1"
    DEFAULT_PORT = 8000
    DEFAULT_PROTOCOL = Protocol.TCP

    def __init__(
        self,
        uri: str,
        *,
        ownership: Ownership = Ownership.FANDANGO_PARTY,
        endpoint_type: EndpointType = EndpointType.CONNECT,
    ):
        party_name, prot, host, port = split_party_spec(uri)
        super().__init__(ownership=ownership, party_name=party_name)
        self.protocol_impl = None

        if prot is None:
            prot = self.DEFAULT_PROTOCOL.value
        protocol = Protocol(prot)
        if host is None:
            host = self.DEFAULT_IP
        info = socket.getaddrinfo(host, None, socket.AF_INET)
        ip = info[0][4][0]
        if isinstance(ip, int):
            raise FandangoValueError(f"Invalid IP address: {ip}")
        if port is None:
            protocol = self.DEFAULT_PORT

        if protocol == Protocol.TCP or protocol == Protocol.UDP:
            self.protocol_impl = UdpTcpProtocolDecorator(
                endpoint_type=endpoint_type,
                protocol_type=protocol,
                ip_type=IpType.IPV4,
                ip=ip,
                port=port,
                party_instance=self,
            )
        else:
            raise FandangoValueError(f"Unsupported protocol: {protocol}")

    def on_send(self, message: DerivationTree, recipient: Optional[str]) -> None:
        if self.protocol_impl is None:
            raise FandangoError("Protocol implementation not initialized.")
        self.protocol_impl.on_send(message, recipient)

    def start(self):
        if self.protocol_impl is None:
            raise FandangoError("Protocol implementation not initialized.")
        self.protocol_impl.start()

    def stop(self):
        if self.protocol_impl is None:
            raise FandangoError("Protocol implementation not initialized.")
        self.protocol_impl.stop()

    @property
    def ip(self):
        if self.protocol_impl is None:
            raise FandangoError("Protocol implementation not initialized.")
        return self.protocol_impl.ip

    @ip.setter
    def ip(self, host: str):
        """Sets the ip for the connection. Applied after a (re)start of the connection party."""
        if self.protocol_impl is None:
            raise FandangoError("Protocol implementation not initialized.")
        info = socket.getaddrinfo(host, None, socket.AF_INET)
        ip = info[0][4][0]
        if isinstance(ip, int):
            raise FandangoValueError(f"Invalid IP address: {ip}")
        self.protocol_impl.ip = ip

    @property
    def port(self):
        if self.protocol_impl is None:
            raise FandangoError("Protocol implementation not initialized.")
        return self.protocol_impl.port

    @port.setter
    def port(self, port: int):
        """Sets the port for the connection. Applied after a (re)start of the connection party."""
        if self.protocol_impl is None:
            raise FandangoError("Protocol implementation not initialized.")
        self.protocol_impl.port = port


class StdOut(FandangoParty):
    """Standard output party for sending messages to stdout. The party can only send messages, but not receive any.
    The party is always owned by Fandango (Ownership.FUZZER), meaning it sends messages generated by Fandango.
    """

    def __init__(self):
        super().__init__(ownership=Ownership.FANDANGO_PARTY)
        self.stream = sys.stdout

    def on_send(self, message: DerivationTree, recipient: Optional[str]):
        """Called by Fandango, when it wants to write a message to StdOut.
        :param message: The message to send.
        :param recipient: The recipient of the message. Only present if the grammar specifies a recipient.
        """
        self.stream.write(message.to_string())


class StdIn(FandangoParty):
    """Standard input party for reading messages from stdin. The party can only receive messages, but not send any.
    The ownership of this party is always Ownership.EXTERNAL, meaning it is an external party.
    """

    def __init__(self):
        super().__init__(ownership=Ownership.EXTERNAL_PARTY)
        self.running = True
        self.stream = sys.stdin
        self.listen_thread = threading.Thread(target=self._listen_loop, daemon=True)
        self.listen_thread.start()

    def _listen_loop(self):
        while self.running:
            rlist, _, _ = select.select([self.stream], [], [], 0.1)
            if rlist:
                read = sys.stdin.readline()
                if read == "":
                    self.running = False
                    break
                self.receive_msg(self.party_name, read)
            else:
                time.sleep(0.1)


class Out(FandangoParty):
    """Standard output party for receiving messages from an external process set using set_program_command(command: str).
    The party can only receive messages, but not send any.
    The ownership of this party is always Ownership.EXTERNAL, meaning it is an external party.
    """

    def __init__(self):
        super().__init__(ownership=Ownership.EXTERNAL_PARTY)
        self.proc = ProcessManager.instance().get_process()
        threading.Thread(target=self._listen_loop, daemon=True).start()

    def _listen_loop(self):
        while True:
            if self.proc.stdout is not None:
                line = self.proc.stdout.read(1)
                self.receive_msg(self.party_name, line)


class In(FandangoParty):
    """Standard input party for sending messages to an external process set using set_program_command(command: str).
    The party can only send messages, but not receive any.
    The ownership of this party is always Ownership.FUZZER, meaning it sends messages generated by Fandango.
    """

    def __init__(self):
        super().__init__(ownership=Ownership.FANDANGO_PARTY)
        self.proc = ProcessManager.instance().get_process()
        self._close_post_transmit = False

    @property
    def close_post_transmit(self) -> bool:
        """Returns whether the stdin of the process should be closed after transmitting a message."""
        return self._close_post_transmit

    @close_post_transmit.setter
    def close_post_transmit(self, value: bool):
        """Sets whether the stdin of the process should be closed after transmitting a message."""
        if self._close_post_transmit == value:
            return
        self._close_post_transmit = value

    def on_send(self, message: DerivationTree, recipient: Optional[str]):
        """Called by Fandango, when it wants to write a message to the external process.
        :param message: The message to send.
        :param recipient: The recipient of the message. Only present if the grammar specifies a recipient.
        """
        if self.proc.stdin is not None:
            self.proc.stdin.write(message.to_string())
            self.proc.stdin.flush()
            if self.close_post_transmit:
                self.proc.stdin.close()


class FandangoIO(object):
    """Singleton class for managing Fandango's input/output operations."""

    _instance: Optional["FandangoIO"] = None

    @classmethod
    def instance(cls) -> "FandangoIO":
        """Returns the singleton instance of FandangoIO. If it does not exist, it creates one.
        Only use this method to access the FandangoIO instance.
        """
        if cls._instance is None:
            FandangoIO()
        assert cls._instance is not None
        return cls._instance

    def __init__(self):
        """Constructor for the FandangoIO class. Singleton! Do not call this method directly. Call instance() instead."""
        assert FandangoIO._instance is None, "FandangoIO singleton already created"
        FandangoIO._instance = self
        self.receive = list[tuple[str, str, str | bytes]]()
        self.parties = dict[str, FandangoParty]()
        self.receive_lock = threading.Lock()

    def add_receive(self, sender: str, receiver: str, message: str | bytes) -> None:
        """Forwards an external, received message to Fandango for processing.
        :param sender: The sender of the message.
        :param receiver: The receiver of the message.
        :param message: The message received from the sender.
        """
        with self.receive_lock:
            self.receive.append((sender, receiver, message))

    def received_msg(self) -> bool:
        """Checks if there are any received messages from external parties."""
        with self.receive_lock:
            return len(self.receive) != 0

    def get_received_msgs(self) -> list[tuple[str, str, str | bytes]]:
        """Returns a list of all received messages from external parties."""
        with self.receive_lock:
            return list(self.receive)

    def clear_received_msg(self, idx: int) -> None:
        """Clears a specific received message by its index."""
        with self.receive_lock:
            del self.receive[idx]

    def clear_received_msgs(self) -> None:
        """Clears all received messages."""
        with self.receive_lock:
            self.receive.clear()

    def clear_by_party(self, party_name: str, to_idx: int) -> None:
        """Clears all received messages from a specific party up to a given index."""

        with self.receive_lock:
            self.receive = [
                (sender, receiver, msg)
                for idx, (sender, receiver, msg) in enumerate(self.receive)
                if not (sender == party_name and idx <= to_idx)
            ]

    def transmit(
        self, sender: str, recipient: Optional[str], message: DerivationTree
    ) -> None:
        """Called by Fandango to transmit a message from a sender to a recipient using the sender's party definition.
        :param sender: The sender of the message. Needs to be equal to the class name of the corresponding party definition.
        :param recipient: The recipient of the message. Only present if the grammar specifies a recipient. Can be used by the party definition to send the message to the correct recipient.
        :param message: The message to send.
        """
        if sender in self.parties.keys():
            self.parties[sender].on_send(message, recipient)


class ProcessManager(object):
    _instance: Optional["ProcessManager"] = None

    def __init__(self):
        """Constructor for the ProcessManager class. Singleton! Do not call this method directly. Call instance() instead."""
        assert (
            ProcessManager._instance is None
        ), "ProcessManager singleton already created"
        ProcessManager._instance = self
        self._command = None
        self.lock = threading.Lock()
        self.proc = None

    @classmethod
    def instance(cls) -> "ProcessManager":
        """Returns the singleton instance of ProcessManager. If it does not exist, it creates one."""
        if cls._instance is None:
            ProcessManager()
        assert cls._instance is not None
        return cls._instance

    def get_process(self) -> subprocess.Popen:
        """Returns the current process if it exists, otherwise starts a new one based on the command set."""
        with self.lock:
            if not self.proc:
                self._start_process()
        if self.proc is None:
            raise FandangoValueError(
                "This spec requires interaction. Use `--party=PARTY` or `fandango talk` with this spec."
            )
        return self.proc

    @property
    def command(self) -> str | list[str] | None:
        """Returns the command to be executed to start the process."""
        return self._command

    def set_command(self, value: str | list[str], text: bool = True):
        """Sets the command to be executed to start the process."""
        assert isinstance(
            value, (str, list)
        ), "Command must be a string or a list of strings"
        with self.lock:
            if self._command == value:
                return
            self._command = value
        self.text = text

    def _start_process(self):
        command = self.command
        if command is None:
            return

        LOGGER.info(f"Starting subprocess with command {command}")
        self.proc = subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=self.text,
        )


def set_program_command(command: str | list[str], text: bool = True):
    """
    Set the command to be executed by the ProcessManager.
    :param command: The command to execute.
    """
    LOGGER.info(f"Setting command {command!r}")
    ProcessManager.instance().set_command(command, text)


if __name__ == "__main__":
    # Some tests for the split_party_spec function
    assert split_party_spec("25") == (None, None, None, 25)
    assert split_party_spec("tcp://localhost:25") == (None, "TCP", "localhost", 25)
    assert split_party_spec("tcp:127.0.0.1:25") == (None, "TCP", "127.0.0.1", 25)
    assert split_party_spec("udp://[::1]:25") == (None, "UDP", "::1", 25)
    assert split_party_spec("tcp://cispa.de:25") == (None, "TCP", "cispa.de", 25)
    assert split_party_spec("SMTP=[::1]:25") == ("SMTP", None, "::1", 25)

    # Demonstrator code to show how to use the classes
    from fandango import Fandango

    SPEC = r"""
    <start> ::= <In:input> <Out:output>
    <input> ::= <string>
    <output> ::= <string>
    <string> ::= r'.*\n'
    where str(<input>) == str(<output>)

    set_program_command("cat")
    """
    fandango = Fandango(SPEC, logging_level=logging.INFO)
    fandango.fuzz()



from typing import TYPE_CHECKING
from collections.abc import Iterator, Sequence
from fandango.language.grammar.has_settings import HasSettings
from fandango.language.grammar.nodes.node import Node, NodeType
from fandango.language.grammar.nodes.terminal import TerminalNode
from fandango.language.symbols.terminal import Terminal
from fandango.language.tree import DerivationTree

if TYPE_CHECKING:
    import fandango


class CharSet(Node):
    def __init__(
        self,
        chars: str,
        grammar_settings: Sequence[HasSettings],
    ):
        self.chars = chars
        super().__init__(NodeType.CHAR_SET, grammar_settings)

    def fuzz(
        self,
        parent: DerivationTree,
        grammar: "fandango.language.grammar.grammar.Grammar",
        max_nodes: int = 100,
        in_message: bool = False,
    ) -> list[DerivationTree]:
        raise NotImplementedError("CharSet fuzzing not implemented")

    def accept(
        self,
        visitor: "fandango.language.grammar.node_visitors.node_visitor.NodeVisitor",
    ):
        return visitor.visitCharSet(self)

    def descendents(
        self, grammar: "fandango.language.grammar.grammar.Grammar"
    ) -> Iterator["Node"]:
        for char in self.chars:
            yield TerminalNode(Terminal(char), self._grammar_settings)

    def format_as_spec(self) -> str:
        return self.chars



from __future__ import annotations
import re
import sys
from typing import TextIO, Optional

# noinspection PyUnresolvedReferences
from fandango.language.parser.FandangoParser import FandangoParser
from antlr4 import Lexer, InputStream, Token
from antlr4.Token import CommonToken

# Current lexer instance, set by the generated lexer code
lexer: Optional["FandangoLexerBase"] = None


class FandangoLexerBase(Lexer):
    NEW_LINE_PATTERN = re.compile("[^\r\n\f]+")
    SPACES_PATTERN = re.compile("[\r\n\f]+")

    def __init__(self, input_: InputStream, output: TextIO = sys.stdout):
        super().__init__(input_, output)
        self.tokens: list[Token] = []
        self.indents: list[int] = []
        self.opened = 0
        self.in_python = 0
        self.in_fstring = False

        # Set the global lexer instance to this one
        global lexer
        lexer = self

    def reset(self):
        self.tokens = []
        self.indents = []
        self.opened = 0
        self.in_python = 0
        self.in_fstring = False
        super().reset()

    def emitToken(self, token: Token):
        self._token = token  # type: ignore[assignment] # library seems to be wrong here?
        self.tokens.append(token)

    def nextToken(self) -> Token:
        # Check if the end-of-file is ahead and there are still some DEDENTS expected.
        if self._input.LA(1) == FandangoParser.EOF and len(self.indents) != 0:
            # Remove any trailing EOF tokens from our buffer.
            self.tokens = [
                token for token in self.tokens if token.type != FandangoParser.EOF
            ]

            # First emit an extra line break that serves as the end of the statement.
            self.emitToken(self.commonToken(FandangoParser.NEWLINE, "\n"))

            # Now emit as much DEDENT tokens as needed.
            while len(self.indents) != 0:
                self.emitToken(self.createDedent())
                self.indents.pop()

            # Put the EOF back on the token stream.
            self.emitToken(self.commonToken(FandangoParser.EOF, "<EOF>"))

        next_ = super().nextToken()
        token = next_ if len(self.tokens) == 0 else self.tokens.pop(0)
        # print(
        #     f"nextToken(): {token.text!r} ({token.type}) at {token.start}..{token.stop}", file=sys.stderr
        # )
        return token

    def createDedent(self) -> Token:
        return self.commonToken(FandangoParser.DEDENT, "")

    def commonToken(self, type_: int, text: str) -> Token:
        stop = self.getCharIndex() - 1
        start = stop if text == "" else stop - len(text) + 1
        token = CommonToken(
            self._tokenFactorySourcePair,
            type_,
            Lexer.DEFAULT_TOKEN_CHANNEL,
            start,
            stop,
        )
        token.text = text
        return token

    @staticmethod
    def get_indentation_count(whitespace: str) -> int:
        count = 0
        for c in whitespace:
            if c == "\t":
                count += 8 - count % 8
            else:
                count += 1
        return count

    def at_start_of_input(self) -> bool:
        return self.getCharIndex() == 0

    def open_brace(self) -> None:
        self.opened += 1

    def close_brace(self) -> None:
        self.opened -= 1

    def python_start(self) -> None:
        self.in_python += 1

    def python_end(self) -> None:
        self.in_python = 0

    # while f-string do not consider the string token
    def fstring_start(self):
        self.in_fstring = True

    def fstring_end(self):
        self.in_fstring = False

    def is_not_fstring(self):
        return not self.in_fstring

    def on_newline(self) -> None:
        new_line = self.NEW_LINE_PATTERN.sub("", self.text)
        spaces = self.SPACES_PATTERN.sub("", self.text)

        next_ = self._input.LA(1)
        next_next = self._input.LA(2)

        if self.opened > 0 or (next_next != -1 and next_ in (10, 13, 35)):
            self.skip()
        else:
            self.emitToken(self.commonToken(FandangoParser.NEWLINE, new_line))
            indent = self.get_indentation_count(spaces)
            previous = 0 if len(self.indents) == 0 else self.indents[-1]

            if indent == previous:
                self.skip()
            elif indent > previous:
                self.indents.append(indent)
                self.emitToken(self.commonToken(FandangoParser.INDENT, spaces))
            else:
                while len(self.indents) > 0 and self.indents[-1] > indent:
                    self.in_python -= 1
                    self.emitToken(self.createDedent())
                    self.indents.pop()


# These are called from the generated lexer code


def at_start_of_input() -> None:
    global lexer
    assert lexer is not None
    lexer.at_start_of_input()


def open_brace() -> None:
    global lexer
    assert lexer is not None
    lexer.open_brace()


def close_brace() -> None:
    global lexer
    assert lexer is not None
    lexer.close_brace()


def python_start() -> None:
    global lexer
    assert lexer is not None
    lexer.python_start()


def python_end() -> None:
    global lexer
    assert lexer is not None
    lexer.python_end()


def on_newline() -> None:
    global lexer
    assert lexer is not None
    lexer.on_newline()


def fstring_start() -> None:
    global lexer
    assert lexer is not None
    return lexer.fstring_start()


def fstring_end() -> None:
    global lexer
    assert lexer is not None
    return lexer.fstring_end()


def is_not_fstring() -> None:
    global lexer
    assert lexer is not None
    return lexer.is_not_fstring()



import random
from collections.abc import Callable, Generator

from fandango.constraints.failing_tree import Comparison, ComparisonSide
from fandango.constraints.failing_tree import FailingTree, BoundsFailingTree
from fandango.constraints.repetition_bounds import RepetitionBoundsConstraint
from fandango.errors import FandangoValueError
from fandango.io.packetforecaster import PacketForecaster
from fandango.language.grammar.grammar import Grammar
from fandango.language.symbols import NonTerminal
from fandango.language.symbols import Slice
from fandango.language.tree import DerivationTree
from fandango.logger import LOGGER


class PopulationManager:
    def __init__(
        self,
        grammar: Grammar,
        start_symbol: str,
        warnings_are_errors: bool = False,
    ):
        self._grammar = grammar
        self._start_symbol = start_symbol
        self._warnings_are_errors = warnings_are_errors

    def _generate_population_entry(self, max_nodes: int):
        return self._grammar.fuzz(self._start_symbol, max_nodes)

    @staticmethod
    def add_unique_individual(
        population: list[DerivationTree],
        candidate: DerivationTree,
        unique_set: set[int],
    ) -> bool:
        """
        Adds individual to the population if it is unique, according to its hash.

        :param population: The population to potentially add the individual to.
        :param candidate: The individual to potentially add to the population.
        :param unique_set: The set of unique individuals.
        :return: True if the individual was added, False otherwise.
        """
        h = hash(candidate)
        if h not in unique_set:
            unique_set.add(h)
            population.append(candidate)
            return True
        return False

    def _is_population_complete(
        self, unique_population: list[DerivationTree], population_size: int
    ) -> bool:
        return len(unique_population) >= population_size

    def refill_population(
        self,
        current_population: list[DerivationTree],
        eval_individual: Callable[
            [DerivationTree],
            Generator[DerivationTree, None, tuple[float, list[FailingTree]]],
        ],
        max_nodes: int,
        target_population_size: int,
    ) -> Generator[DerivationTree, None, None]:
        """
        Refills the population with unique individuals in place.

        Does not deduplicate the current population.

        If after 10 times the difference between the current population size and the target population size
        the required population size is still not met, a warning is logged and the incomplete population is returned.

        :param current_population: The current population of individuals.
        :param eval_individual: The function to evaluate the fitness of an individual.
        :param max_nodes: The maximum number of nodes in an individual.
        :param target_population_size: The target size of the population.
        :return: A generator that yields solutions. The population is modified in place.
        """
        unique_hashes = {hash(ind) for ind in current_population}
        attempts = 0
        max_attempts = (target_population_size - len(current_population)) * 10

        while (
            not self._is_population_complete(current_population, target_population_size)
            and attempts < max_attempts
        ):
            individual = self._generate_population_entry(max_nodes)
            _fitness, failing_trees = yield from eval_individual(individual)
            candidate, _fixes_made = self.fix_individual(
                individual,
                failing_trees,
            )
            _new_fitness, _new_failing_trees = yield from eval_individual(candidate)
            if not PopulationManager.add_unique_individual(
                current_population, candidate, unique_hashes
            ):
                attempts += 1

        if not self._is_population_complete(current_population, target_population_size):
            LOGGER.warning(
                f"Could not generate a full population of unique individuals. Population size reduced to {len(current_population)}."
            )

    def fix_individual(
        self,
        individual: DerivationTree,
        failing_trees: list[FailingTree],
    ) -> tuple[DerivationTree, int]:
        fixes_made = 0
        replacements: list[tuple[DerivationTree, DerivationTree]] = list()

        allow_repetition_full_delete = (
            len(
                list(
                    filter(
                        lambda x: not isinstance(x.cause, RepetitionBoundsConstraint),
                        failing_trees,
                    )
                )
            )
            == 0
        )
        # We only allow BoundsConstraints to delete all iterations of a repetition if all other non-boundconstraints constraints are satisfied.
        # Otherwise, we would lose the reference point to re-add the repetitions in the tree, which might be needed,
        # if the referenced length field changes its value.
        # This is a workaround for the fact that we cannot delete all repetitions in a tree, if there

        for failing_tree in failing_trees:
            if failing_tree.tree.read_only:
                continue

            if isinstance(failing_tree, BoundsFailingTree):
                assert isinstance(failing_tree.cause, RepetitionBoundsConstraint)
                bounds_constraint: RepetitionBoundsConstraint = failing_tree.cause
                replacements.extend(
                    bounds_constraint.fix_individual(
                        self._grammar,
                        failing_tree,
                        allow_repetition_full_delete=allow_repetition_full_delete,
                    )
                )
                continue

            for operator, value, side in failing_tree.suggestions:
                if operator == Comparison.EQUAL and side == ComparisonSide.LEFT:
                    # LOGGER.debug(f"Parsing {value} into {failing_tree.tree.symbol.symbol!s}")
                    symbol = failing_tree.tree.symbol
                    if isinstance(value, DerivationTree) and symbol == value.symbol:
                        suggested_tree = value.deepcopy(
                            copy_children=True, copy_params=False, copy_parent=False
                        )
                        suggested_tree.set_all_read_only(False)
                    elif isinstance(symbol, NonTerminal):
                        suggested_tree = self._grammar.parse(value, start=symbol)
                    elif isinstance(symbol, Slice):
                        # slices don't have a symbol associated with them — I think
                        suggested_tree = self._grammar.parse(value, start="")
                    if suggested_tree is None:
                        continue
                    replacements.append((failing_tree.tree, suggested_tree))
                    fixes_made += 1
        if len(replacements) > 0:
            # Prevent circular replacements
            # deleted = set()
            # for value in set(replacements.values()):
            #    if value in deleted:
            #        continue
            #    if value in replacements.keys():
            #        if replacements[value] not in replacements.keys():
            #            deleted.add(replacements[value])
            #            del replacements[value]
            #            continue
            #        if random.random() < 0.5:
            #            deleted.add(replacements[value])
            #            del replacements[value]
            #        else:
            #            deleted.add(replacements[replacements[value]])
            #            del replacements[replacements[value]]

            individual = individual.replace_multiple(self._grammar, replacements)
        return individual, fixes_made


class IoPopulationManager(PopulationManager):
    def __init__(
        self,
        grammar: Grammar,
        start_symbol: str,
        warnings_are_errors: bool = False,
    ):
        super().__init__(grammar, start_symbol, warnings_are_errors)
        self._prev_packet_idx = 0
        self.fuzzable_packets: list[PacketForecaster.ForcastingPacket] | None = None

    def _generate_population_entry(self, max_nodes: int):
        if self.fuzzable_packets is None or len(self.fuzzable_packets) == 0:
            return DerivationTree(NonTerminal(self._start_symbol))

        current_idx = (self._prev_packet_idx + 1) % len(self.fuzzable_packets)
        current_pck = random.choice(self.fuzzable_packets)
        mounting_option = random.choice(list(current_pck.paths))

        tree = self._grammar.collapse(mounting_option.tree)
        if tree is None:
            raise FandangoValueError(
                f"Could not collapse tree for {mounting_option.path} in packet {current_pck.node}"
            )
        tree.set_all_read_only(True)
        dummy = DerivationTree(NonTerminal("<hookin>"))
        tree.append(mounting_option.path[1:], dummy)

        fuzz_point = dummy.parent
        assert fuzz_point is not None
        fuzz_point.set_children(fuzz_point.children[:-1])
        current_pck.node.fuzz(fuzz_point, self._grammar, max_nodes)

        self._prev_packet_idx = current_idx
        return tree



from copy import copy
from typing import Any, Optional
from fandango.constraints.constraint_visitor import ConstraintVisitor
from fandango.constraints.constraint import Constraint
from fandango.constraints.fitness import ConstraintFitness, FailingTree
from fandango.language.symbols.non_terminal import NonTerminal
from fandango.language.tree import DerivationTree
from fandango.logger import print_exception


class ExpressionConstraint(Constraint):
    """
    Represents a python expression constraint that can be used for fitness evaluation.
    """

    def __init__(self, expression: str, *args, **kwargs):
        """
        Initializes the expression constraint with the given expression.
        :param str expression: The expression to evaluate.
        :param args: Additional arguments.
        :param kwargs: Additional keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.expression = expression

    def fitness(
        self,
        tree: DerivationTree,
        scope: Optional[dict[NonTerminal, DerivationTree]] = None,
        population: Optional[list[DerivationTree]] = None,
        local_variables: Optional[dict[str, Any]] = None,
    ) -> ConstraintFitness:
        """
        Calculate the fitness of the tree based on whether the given expression evaluates to True.
        :param DerivationTree tree: The tree to evaluate.
        :param Optional[dict[str, DerivationTree]] scope: The scope of the tree.
        :param Optional[list[DerivationTree]] population: The population of trees.
        :param Optional[dict[str, Any]] local_variables: Local variables to use in the evaluation.
        :return ConstraintFitness: The fitness of the tree.
        """
        tree_hash = self.get_hash(tree, scope, population, local_variables)
        # If the fitness has already been calculated, return the cached value
        if tree_hash in self.cache:
            return copy(self.cache[tree_hash])
        # Initialize the fitness values
        solved = 0
        total = 0
        failing_trees = []
        # If the tree is None, the fitness is 0
        if tree is None:
            return ConstraintFitness(0, 0, False)
        has_combinations = False
        # Iterate over all combinations of the tree and the scope
        for combination in self.combinations(tree, scope, population):
            has_combinations = True
            # Update the local variables to initialize the placeholders with the values of the combination
            local_vars = self.local_variables.copy()
            if local_variables:
                local_vars.update(local_variables)
            local_vars.update(
                {name: container.evaluate() for name, container in combination}
            )
            try:
                result = self.eval(self.expression, self.global_variables, local_vars)
                # Commented this out for now, as `None` is a valid result
                # of functions such as `re.match()` -- AZ
                # if result is None:
                #     return ConstraintFitness(1, 1, True)
                if result:
                    solved += 1
                else:
                    # If the expression evaluates to False, add the failing trees to the list
                    for _, container in combination:
                        for node in container.get_trees():
                            if node not in failing_trees:
                                failing_trees.append(node)
            except Exception as e:
                print_exception(e, f"Evaluation failed: {self.expression}")

            total += 1
        # If there are no combinations, the fitness is perfect
        if not has_combinations:
            solved += 1
            total += 1
        # Create the fitness object
        fitness = ConstraintFitness(
            solved,
            total,
            solved == total,
            failing_trees=[FailingTree(t, self) for t in failing_trees],
        )
        # Cache the fitness
        self.cache[tree_hash] = fitness
        return fitness

    def format_as_spec(self) -> str:
        representation = self.expression
        for identifier in self.searches:
            representation = representation.replace(
                identifier, self.searches[identifier].format_as_spec()
            )
        return representation

    def accept(self, visitor: "ConstraintVisitor"):
        """
        Accepts a visitor to traverse the constraint structure.
        :param ConstraintVisitor visitor: The visitor to accept.
        """
        visitor.visit_expression_constraint(self)



import random
from typing import Any, Callable
from fandango.language.grammar.grammar import Grammar
from fandango.language.grammar.grammar_settings import GrammarSetting
from fandango.language.grammar.nodes.node import Node, NODE_SETTINGS_DEFAULTS
from fandango.language.grammar.nodes.non_terminal import NonTerminalNode
from fandango.language.grammar.nodes.terminal import TerminalNode
from fandango.language.parse import parse

FLOAT_KEY = next(
    k for k in NODE_SETTINGS_DEFAULTS if isinstance(NODE_SETTINGS_DEFAULTS[k], float)
)
INT_KEY = next(
    k for k in NODE_SETTINGS_DEFAULTS if isinstance(NODE_SETTINGS_DEFAULTS[k], int)
)


def get_grammar(
    selector_probability_pairs: list[tuple[str, dict[str, Any]]],
) -> Grammar:
    raw_spec = f"""
<start> ::= "a" | <b>*
<b> ::= "b"
"""

    for selector, settings in selector_probability_pairs:
        # try both '=' and ' ' as separators
        oddness = random.choice([0, 1])
        kv_pairs = " ".join(
            [
                f"{k}{'=' if i % 2 == oddness else ' '}{v}"
                for i, (k, v) in enumerate(settings.items())
            ]
        )
        raw_spec += f"""setting {selector} {kv_pairs}
"""
    grammar, constraints = parse(raw_spec, use_cache=False, use_stdlib=False)
    assert grammar is not None
    assert constraints is not None and len(constraints) == 0
    return grammar


def check_node(
    node: Node,
    float_key_on_set: float,
    int_key_on_set: int | None,
    match_lambda: Callable[[Node], bool],
):
    if match_lambda(node):
        assert node._settings._settings[FLOAT_KEY] == float_key_on_set
        if int_key_on_set is not None:
            assert node._settings._settings[INT_KEY] == int_key_on_set
        else:
            assert INT_KEY not in node._settings._settings
    else:
        assert FLOAT_KEY not in node._settings._settings
        assert INT_KEY not in node._settings._settings
    for n in node.children():
        check_node(n, float_key_on_set, int_key_on_set, match_lambda)


def test_can_parse_grammar_settings():
    grammar = get_grammar([("<start>", {FLOAT_KEY: 0.5})])
    gss = grammar._grammar_settings
    assert len(gss) == 1
    gs = gss[0]
    assert isinstance(gs, GrammarSetting)
    assert gs._selector == "<start>"
    ns = gs._node_settings
    assert ns._settings[FLOAT_KEY] == 0.5
    assert ns.get(FLOAT_KEY) == 0.5
    assert INT_KEY not in ns._settings
    assert ns.get(INT_KEY) == NODE_SETTINGS_DEFAULTS[INT_KEY]

    def match_start(node: Node) -> bool:
        return isinstance(node, NonTerminalNode) and node.symbol.name() == "<start>"

    for node in grammar.rules.values():
        check_node(node, 0.5, None, match_start)

    for nt in grammar.rules.keys():
        check_node(NonTerminalNode(nt, gss), 0.5, None, match_start)


def test_assign_to_all_nonterminals():
    grammar = get_grammar([("all_with_type(NonTerminalNode)", {FLOAT_KEY: 0.5})])
    gss = grammar._grammar_settings
    assert len(gss) == 1
    gs = gss[0]
    assert isinstance(gs, GrammarSetting)
    assert gs._selector == "all_with_type(NonTerminalNode)"
    ns = gs._node_settings
    assert ns._settings[FLOAT_KEY] == 0.5
    assert ns.get(FLOAT_KEY) == 0.5
    assert INT_KEY not in ns._settings
    assert ns.get(INT_KEY) == NODE_SETTINGS_DEFAULTS[INT_KEY]

    def match_nts(node: Node) -> bool:
        return isinstance(node, NonTerminalNode)

    for node in grammar.rules.values():
        check_node(node, 0.5, None, match_nts)

    for nt in grammar.rules.keys():
        check_node(NonTerminalNode(nt, gss), 0.5, None, match_nts)


def test_assign_to_all_terminals():
    grammar = get_grammar([("all_with_type(TerminalNode)", {FLOAT_KEY: 0.5})])
    gss = grammar._grammar_settings
    assert len(gss) == 1
    gs = gss[0]
    assert isinstance(gs, GrammarSetting)
    assert gs._selector == "all_with_type(TerminalNode)"
    ns = gs._node_settings
    assert ns._settings[FLOAT_KEY] == 0.5
    assert ns.get(FLOAT_KEY) == 0.5
    assert INT_KEY not in ns._settings
    assert ns.get(INT_KEY) == NODE_SETTINGS_DEFAULTS[INT_KEY]

    def match_ts(node: Node) -> bool:
        return isinstance(node, TerminalNode)

    for node in grammar.rules.values():
        check_node(node, 0.5, None, match_ts)

    for nt in grammar.rules.keys():
        check_node(NonTerminalNode(nt, gss), 0.5, None, match_ts)


def test_assign_to_all():
    grammar = get_grammar([("*", {FLOAT_KEY: 0.5})])
    gss = grammar._grammar_settings
    assert len(gss) == 1
    gs = gss[0]
    assert isinstance(gs, GrammarSetting)
    assert gs._selector == "*"
    assert gs._node_settings._settings[FLOAT_KEY] == 0.5
    assert gs._node_settings.get(FLOAT_KEY) == 0.5
    assert INT_KEY not in gs._node_settings._settings
    assert gs._node_settings.get(INT_KEY) == NODE_SETTINGS_DEFAULTS[INT_KEY]

    def match_all(node: Node) -> bool:
        return True

    for node in grammar.rules.values():
        check_node(node, 0.5, None, match_all)

    for nt in grammar.rules.keys():
        check_node(NonTerminalNode(nt, gss), 0.5, None, match_all)


def test_multiple_settings_lines():
    grammar = get_grammar(
        [
            ("all_with_type(NonTerminalNode)", {FLOAT_KEY: 0.5}),
            ("all_with_type(TerminalNode)", {FLOAT_KEY: 0.5}),
        ]
    )
    gss = grammar._grammar_settings
    assert len(gss) == 2
    for gs in gss:
        assert isinstance(gs, GrammarSetting)
        assert (
            gs._selector == "all_with_type(NonTerminalNode)"
            or gs._selector == "all_with_type(TerminalNode)"
        )
        assert gs._node_settings._settings[FLOAT_KEY] == 0.5
        assert INT_KEY not in gs._node_settings._settings

    def match_ts_and_nts(node: Node) -> bool:
        return isinstance(node, NonTerminalNode) or isinstance(node, TerminalNode)

    for node in grammar.rules.values():
        check_node(node, 0.5, None, match_ts_and_nts)

    for nt in grammar.rules.keys():
        check_node(NonTerminalNode(nt, gss), 0.5, None, match_ts_and_nts)


def test_multiple_key_value_pairs():
    grammar = get_grammar(
        [
            (
                "all_with_type(NonTerminalNode)",
                {FLOAT_KEY: 0.5, INT_KEY: 3},
            )
        ]
    )
    gss = grammar._grammar_settings
    assert len(gss) == 1
    gs = gss[0]
    assert isinstance(gs, GrammarSetting)
    assert gs._selector == "all_with_type(NonTerminalNode)"
    ns = gs._node_settings
    assert ns._settings[FLOAT_KEY] == 0.5
    assert ns._settings[INT_KEY] == 3
    assert ns.get(FLOAT_KEY) == 0.5
    assert ns.get(INT_KEY) == 3

    def match_nts(node: Node) -> bool:
        return isinstance(node, NonTerminalNode)

    for node in grammar.rules.values():
        check_node(node, 0.5, 3, match_nts)

    for nt in grammar.rules.keys():
        check_node(NonTerminalNode(nt, gss), 0.5, 3, match_nts)



# from https://microsoft.github.io/language-server-protocol/specifications/lsp/3.17/specification/#semanticTokenTypes

import enum

from antlr4 import Token


class SemanticTokenTypes(enum.Enum):
    namespace = "namespace"
    # /**
    #  * Represents a generic type. Acts as a fallback for types which
    #  * can't be mapped to a specific type like class or enum.
    #  */
    type = "type"
    _class = "class"
    enum = "enum"
    interface = "interface"
    struct = "struct"
    typeParameter = "typeParameter"
    parameter = "parameter"
    variable = "variable"
    property = "property"
    enumMember = "enumMember"
    event = "event"
    function = "function"
    method = "method"
    macro = "macro"
    keyword = "keyword"
    modifier = "modifier"
    comment = "comment"
    string = "string"
    number = "number"
    regexp = "regexp"
    operator = "operator"
    decorator = "decorator"

    @classmethod
    def values(cls) -> list[str]:
        """Get the values of the semantic token types.

        :return: The values of the semantic token types
        """
        return [e.value for e in cls]

    @classmethod
    def from_token(cls, token: Token) -> "SemanticTokenTypes":
        """Get the semantic token type for a given token.

        :param token: The token to get the semantic token type for
        :return: The semantic token type
        """
        match token.type:
            case "NAME":
                return SemanticTokenTypes.variable
            case "STRING_LITERAL" | "BYTES_LITERAL":
                return SemanticTokenTypes.string
            case (
                "DECIMAL_INTEGER"
                | "OCT_INTEGER"
                | "HEX_INTEGER"
                | "BIN_INTEGER"
                | "FLOAT_NUMBER"
                | "IMAG_NUMBER"
            ):
                return SemanticTokenTypes.number
            case "GRAMMAR_ASSIGN" | "EXPR_ASSIGN" | "OR_OP":
                return SemanticTokenTypes.operator
            case "SKIP_":
                return SemanticTokenTypes.comment
            case _:
                return SemanticTokenTypes.type

    @classmethod
    def from_token_as_number(cls, token: Token) -> int:
        """Get the zero-based index of an enum member in its enum class.

        :param token: The token to get the index for
        :return: The zero-based index of the enum member
        """
        return get_enum_index(cls.from_token(token))


class SemanticTokenModifiers(enum.Enum):
    declaration = "declaration"
    definition = "definition"
    readonly = "readonly"
    static = "static"
    deprecated = "deprecated"
    abstract = "abstract"
    _async = "async"
    modification = "modification"
    documentation = "documentation"
    defaultLibrary = "defaultLibrary"

    @classmethod
    def values(cls) -> list[str]:
        """Get the values of the semantic token modifiers.

        :return: The values of the semantic token modifiers
        """
        return [e.value for e in cls]

    @classmethod
    def from_token(cls, token: Token) -> list["SemanticTokenModifiers"]:
        """Get the semantic token modifiers for a given token.

        :param token: The token to get the semantic token modifiers for
        :return: The semantic token modifiers
        """
        return []  # not currently implemented

    @classmethod
    def from_token_as_number(cls, token: Token) -> int:
        """
        Get the bitmask of the semantic token modifiers for a given token.

        The bitmask is the result of XORing together the indices of the enum members in the enum class.

        :param token: The token to get the index for
        :return: The zero-based index of the enum member
        """
        ints = [get_enum_index(m) for m in cls.from_token(token)]
        res = 0
        for i in ints:
            res |= 1 << i
        return res


def get_enum_index(enum_member: enum.Enum) -> int:
    """Get the zero-based index of an enum member in its enum class.

    :param enum_member: The enum member to get the index for
    :return: The zero-based index of the enum member
    """
    return list(type(enum_member)).index(enum_member)



from copy import copy
from fandango.constraints.constraint_visitor import ConstraintVisitor
from typing import Any, Optional
from fandango import DerivationTree
from fandango.constraints.constraint import Constraint
from fandango.constraints.fitness import ConstraintFitness
from fandango.language.symbols.non_terminal import NonTerminal
from fandango.language.tree import DerivationTree


class ImplicationConstraint(Constraint):
    """
    Represents an implication constraint that can be used for fitness evaluation.
    """

    def __init__(self, antecedent: Constraint, consequent: Constraint, *args, **kwargs):
        """
        Initializes the implication constraint with the given antecedent and consequent.
        :param Constraint antecedent: The antecedent of the implication.
        :param Constraint consequent: The consequent of the implication.
        """
        super().__init__(*args, **kwargs)
        self.antecedent = antecedent
        self.consequent = consequent

    def fitness(
        self,
        tree: DerivationTree,
        scope: Optional[dict[NonTerminal, DerivationTree]] = None,
        population: Optional[list[DerivationTree]] = None,
        local_variables: Optional[dict[str, Any]] = None,
    ) -> ConstraintFitness:
        """
        Calculate the fitness of the tree based on the given implication.
        :param DerivationTree tree: The tree to evaluate.
        :param Optional[dict[str, DerivationTree]] scope: The scope of the tree.
        :param Optional[list[DerivationTree]] population: The population of trees.
        :param Optional[dict[str, Any]] local_variables: Local variables to use in the evaluation.
        :return ConstraintFitness: The fitness of the tree.
        """
        tree_hash = self.get_hash(tree, scope, population, local_variables)
        # If the fitness has already been calculated, return the cached value
        if tree_hash in self.cache:
            return copy(self.cache[tree_hash])
        # Evaluate the antecedent
        antecedent_fitness = self.antecedent.fitness(
            tree, scope, population, local_variables
        )
        if antecedent_fitness.success:
            # If the antecedent is true, evaluate the consequent
            fitness = copy(
                self.consequent.fitness(tree, scope, population, local_variables)
            )
            fitness.total += 1
            if fitness.success:
                fitness.solved += 1
        else:
            # If the antecedent is false, the fitness is perfect
            fitness = ConstraintFitness(
                1,
                1,
                True,
            )
        # Cache the fitness
        self.cache[tree_hash] = fitness
        return fitness

    def format_as_spec(self) -> str:
        return f"({self.antecedent.format_as_spec()} -> {self.consequent.format_as_spec()})"

    def accept(self, visitor: "ConstraintVisitor"):
        """
        Accepts a visitor to traverse the constraint structure.
        :param ConstraintVisitor visitor: The visitor to accept.
        """
        visitor.visit_implication_constraint(self)
        if visitor.do_continue(self):
            self.antecedent.accept(visitor)
            self.consequent.accept(visitor)



from typing import TYPE_CHECKING
from collections.abc import Sequence
from fandango.language.grammar.has_settings import HasSettings
from fandango.language.grammar.node_visitors.node_visitor import NodeVisitor
from fandango.language.grammar.nodes.alternative import Alternative
from fandango.language.grammar.nodes.char_set import CharSet
from fandango.language.grammar.nodes.concatenation import Concatenation
from fandango.language.grammar.nodes.node import Node
from fandango.language.grammar.nodes.non_terminal import NonTerminalNode
from fandango.language.grammar.nodes.repetition import Repetition, Star, Plus, Option
from fandango.language.grammar.nodes.terminal import TerminalNode
from fandango.language.symbols import NonTerminal, Terminal, Slice
from fandango.language.symbols.symbol import Symbol

if TYPE_CHECKING:
    import fandango


class Disambiguator(NodeVisitor):
    def __init__(
        self,
        grammar: "fandango.language.grammar.grammar.Grammar",
        grammar_settings: Sequence[HasSettings],
    ):
        self.known_disambiguations: dict[
            Node,
            dict[tuple[NonTerminal | Terminal | Slice, ...], list[tuple[Node, ...]]],
        ] = {}
        self.grammar = grammar
        self._grammar_settings = grammar_settings

    def visit(
        self, node: Node
    ) -> dict[tuple[NonTerminal | Terminal | Slice, ...], list[tuple[Node, ...]]]:
        if node in self.known_disambiguations:
            return self.known_disambiguations[node]
        result = super().visit(node)
        self.known_disambiguations[node] = result
        return result

    def visitAlternative(
        self, node: Alternative
    ) -> dict[tuple[NonTerminal | Terminal | Slice, ...], list[tuple[Node, ...]]]:
        child_endpoints: dict[
            tuple[NonTerminal | Terminal | Slice, ...], list[tuple[Node, ...]]
        ] = {}
        for child in node.children():
            endpoints = self.visit(child)
            for children in endpoints:
                # prepend the alternative to all paths
                if children not in child_endpoints:
                    child_endpoints[children] = []
                # join observed paths (these are impossible to disambiguate)
                child_endpoints[children].extend(
                    (node,) + path for path in endpoints[children]
                )

        return child_endpoints

    def visitConcatenation(
        self, node: Concatenation
    ) -> dict[tuple[NonTerminal | Terminal | Slice, ...], list[tuple[Node, ...]]]:
        child_endpoints: dict[
            tuple[NonTerminal | Terminal | Slice, ...], list[tuple[Node, ...]]
        ] = {(): []}
        for child in node.children():
            next_endpoints: dict[
                tuple[NonTerminal | Terminal | Slice, ...], list[tuple[Node, ...]]
            ] = {}
            endpoints = self.visit(child)
            for children in endpoints:
                for existing in child_endpoints:
                    concatenation = existing + children
                    if concatenation not in next_endpoints:
                        next_endpoints[concatenation] = []
                    next_endpoints[concatenation].extend(child_endpoints[existing])
                    next_endpoints[concatenation].extend(endpoints[children])
            child_endpoints = next_endpoints

        return {
            children: [(node,) + path for path in child_endpoints[children]]
            for children in child_endpoints
        }

    def visitRepetition(
        self, node: Repetition
    ) -> dict[tuple[NonTerminal | Terminal | Slice, ...], list[tuple[Node, ...]]]:
        # repetitions are alternatives over concatenations
        implicit_alternative = next(node.descendents(self.grammar))
        return self.visit(implicit_alternative)

    def visitStar(
        self, node: Star
    ) -> dict[tuple[NonTerminal | Terminal | Slice, ...], list[tuple[Node, ...]]]:
        return self.visitRepetition(node)

    def visitPlus(
        self, node: Plus
    ) -> dict[tuple[NonTerminal | Terminal | Slice, ...], list[tuple[Node, ...]]]:
        return self.visitRepetition(node)

    def visitOption(
        self, node: Option
    ) -> dict[tuple[NonTerminal | Terminal | Slice, ...], list[tuple[Node, ...]]]:
        implicit_alternative = Alternative(
            [
                Concatenation([], self._grammar_settings),
                Concatenation([node.node], self._grammar_settings),
            ],
            self._grammar_settings,
        )
        return self.visit(implicit_alternative)

    def visitNonTerminalNode(
        self, node: NonTerminalNode
    ) -> dict[tuple[Symbol, ...], list[tuple[Node, ...]]]:
        return {(node.symbol,): [(node,)]}

    def visitTerminalNode(
        self, node: TerminalNode
    ) -> dict[tuple[Symbol, ...], list[tuple[Node, ...]]]:
        return {(node.symbol,): [(node,)]}

    def visitCharSet(
        self, node: CharSet
    ) -> dict[tuple[Symbol, ...], list[tuple[Node, ...]]]:
        return {
            (Terminal(c),): [(node, TerminalNode(Terminal(c), self._grammar_settings))]
            for c in node.chars
        }



#!/usr/bin/env pytest

import itertools
import random
import unittest

from fandango.evolution.algorithm import Fandango
from fandango.language.symbols import NonTerminal
from fandango.language.parse import parse
from fandango.language.tree import DerivationTree
from .utils import RESOURCES_ROOT


class ConstraintTest(unittest.TestCase):
    def count_g_params(self, tree: DerivationTree) -> int:
        count = 0
        if len(tree.sources) > 0:
            count += 1
        assert tree.children is not None
        for child in tree.children:
            count += self.count_g_params(child)
        for child in tree.sources:
            count += self.count_g_params(child)
        return count

    def test_generate_k_paths(self):
        with open(RESOURCES_ROOT / "grammar.fan", "r") as file:
            grammar, _ = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None

        k_paths = grammar._generate_all_k_paths(3)
        print(len(k_paths))

    def test_derivation_k_paths(self):
        with open(RESOURCES_ROOT / "grammar.fan", "r") as file:
            grammar, _ = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None

        random.seed(0)
        tree = grammar.fuzz()
        print([t.symbol.format_as_spec() for t in tree.flatten()])

    def test_parse(self):
        with open(RESOURCES_ROOT / "grammar.fan", "r") as file:
            grammar, _ = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None
        tree = grammar.parse("aabb")
        assert tree is not None
        for path in grammar.traverse_derivation(tree):
            for node in path:
                print(node.format_as_spec())

    @staticmethod
    def get_solutions(grammar, constraints, desired_solutions=1):
        fandango = Fandango(grammar=grammar, constraints=constraints)
        return list(itertools.islice(fandango.generate(), desired_solutions))

    def test_generators(self):
        with open(RESOURCES_ROOT / "bar.fan", "r") as file:
            grammar, constraints = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None

        # grammar produces 1 output
        actual = self.get_solutions(grammar, constraints, desired_solutions=1)
        self.assertEqual(len(actual), 1)
        res = actual[0]
        self.assertIsInstance(res, DerivationTree)
        self.assertEqual(str(res.value()), "bar")

    def test_nested_generators(self):
        with open(RESOURCES_ROOT / "nested_grammar_parameters.fan", "r") as file:
            grammar, c = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None

        for solution in self.get_solutions(grammar, c, desired_solutions=10):
            self.assertEqual(self.count_g_params(solution), 4)
            converted_inner = solution.children[0].sources[0]
            self.assertEqual(self.count_g_params(converted_inner), 3)
            dummy_inner_2 = converted_inner.children[0].sources[0]
            self.assertEqual(self.count_g_params(dummy_inner_2), 2)
            dummy_inner = dummy_inner_2.children[0].sources[0]
            self.assertEqual(self.count_g_params(dummy_inner), 1)
            source_nr = dummy_inner.children[0].children[1].sources[0]
            self.assertEqual(self.count_g_params(source_nr), 0)

    def test_repetitions(self):
        with open(RESOURCES_ROOT / "repetitions.fan", "r") as file:
            grammar, c = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None

        expected = ["aaa"]
        actual = self.get_solutions(grammar, c, desired_solutions=1)

        self.assertListEqual(expected, actual)

    def test_repetitions_slice(self):
        with open(RESOURCES_ROOT / "slicing.fan", "r") as file:
            grammar, c = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None

        # grammar produces 7 outputs
        solutions = self.get_solutions(grammar, c, desired_solutions=7)
        for solution in solutions:
            self.assertGreaterEqual(len(str(solution)), 3)
            self.assertLessEqual(len(str(solution)), 10)

    def test_repetition_min(self):
        with open(RESOURCES_ROOT / "min_reps.fan", "r") as file:
            grammar, c = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None

        solutions = self.get_solutions(grammar, c, desired_solutions=20)
        for solution in solutions:
            self.assertGreaterEqual(len(str(solution)), 3)

    def test_repetition_computed(self):
        with open(RESOURCES_ROOT / "dynamic_repetition.fan", "r") as file:
            grammar, c = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None
        solutions = self.get_solutions(grammar, c, desired_solutions=10)
        for solution in solutions:
            len_outer = solution.children[0].to_int()
            self.assertEqual(len_outer, len(solution.children) - 3)
            for tree in solution.children[2:-1]:
                len_inner = tree.children[0].to_int()
                self.assertEqual(len_inner, len(tree.children) - 1)

    def test_repetition_computed_b(self):
        with open(RESOURCES_ROOT / "dynamic_repetition_2.fan", "r") as file:
            grammar, c = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None
        solutions = self.get_solutions(grammar, c)
        for solution in solutions:
            len_a = solution.children[0].to_int()
            assert len_a is not None
            self.assertLessEqual(len_a + 2, len(solution.children))
            for child in solution.children[1 : len_a + 1]:
                self.assertTrue(child.symbol == NonTerminal("<a>"))
            len_b = solution.children[len_a + 1]
            assert len_b is not None
            self.assertTrue(len_b.symbol == NonTerminal("<len_b>"))
            len_b = len_b.to_int()
            assert len_b is not None
            self.assertEqual(len_a + len_b + 2, len(solution.children))
            for child in solution.children[len_a + 4 :]:
                self.assertTrue(child.symbol == NonTerminal("<b>"))

    def test_generator_redefinition(self):
        with open(RESOURCES_ROOT / "generator_remove.fan", "r") as file:
            grammar, c = parse(file, use_stdlib=True, use_cache=False)
            assert grammar is not None
        solutions = self.get_solutions(grammar, c, desired_solutions=10)
        for solution in solutions:
            self.assertNotEqual(str(solution.value()), "10")

    def test_max_nodes(self):
        with open(RESOURCES_ROOT / "gen_number.fan", "r") as file:
            grammar, c = parse(file, use_cache=False, use_stdlib=True)
        solution = self.get_solutions(grammar, c, desired_solutions=10)
        for sol in solution:
            s = str(sol).split(".")
            self.assertEqual(s[0], "a" * 50)
            self.assertTrue(len(s[1]) >= 10)



# file generated by setuptools_scm
# don't change, don't track in version control
TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Union

    VERSION_TUPLE = tuple[Union[int, str], ...]
else:
    VERSION_TUPLE = object

version: str
__version__: str
__version_tuple__: VERSION_TUPLE
version_tuple: VERSION_TUPLE

__version__ = version = "1.0.0.post348+g061673d.d20241214"
__version_tuple__ = version_tuple = (1, 0, 0, "g061673d.d20241214")



import pytest
from fandango.errors import FandangoConversionError
from fandango.language.symbols.terminal import Terminal
from fandango.language.tree import DerivationTree
from fandango.language.tree_value import (
    DIRECT_ACCESS_METHODS_BASE_TO_FIRST_ARG_TYPE,
    DIRECT_ACCESS_METHODS_BASE_TO_UNDERLYING_TYPE,
    TreeValue,
    trailing_bits_to_int,
)
from fandango.language.tree_value import TreeValue, TreeValueType, trailing_bits_to_int


A_BITS = [int(bit) for bit in f"{ord('a'):08b}"]
ONE_BITS = [int(bit) for bit in f"{ord('1'):08b}"]


def test_trailing_bits_to_int():
    assert trailing_bits_to_int([]) == 0
    assert trailing_bits_to_int([0]) == 0
    assert trailing_bits_to_int([1]) == 1
    assert trailing_bits_to_int([0, 1]) == 1
    assert trailing_bits_to_int([1, 0]) == 2
    assert trailing_bits_to_int([1, 1]) == 3
    assert trailing_bits_to_int([1] + [0] * 16) == 65536


@pytest.mark.parametrize(
    "value", ["Hello, World!", "", "💃", b"", b"Hello, World!", 0, 1]
)
def test_tree_value_create(value):
    type_ = type(value)
    tree_value = TreeValue(value)
    assert type_(tree_value) == value


def test_empty_convertors():
    tree_value = TreeValue.empty()
    assert str(tree_value) == ""
    assert bytes(tree_value) == b""
    assert int(tree_value) == 0
    assert tree_value.to_bits() == ""


def test_append_to_empty():
    tree_value = TreeValue.empty()
    tree_value = tree_value.append(TreeValue(1))
    with pytest.raises(ValueError):
        str(tree_value)
    with pytest.raises(ValueError):
        bytes(tree_value)
    assert int(tree_value) == 1
    assert tree_value.to_bits() == "1"

    tree_value = TreeValue.empty()
    tree_value = tree_value.append(TreeValue("a"))
    assert str(tree_value) == "a"
    assert bytes(tree_value) == b"a"
    with pytest.raises(ValueError):
        assert int(tree_value) == 1


def test_append_with_empty():
    tree_value = TreeValue(1)
    tree_value = tree_value.append(TreeValue.empty())
    with pytest.raises(ValueError):
        str(tree_value)
    with pytest.raises(ValueError):
        bytes(tree_value)
    assert int(tree_value) == 1
    assert tree_value.to_bits() == "1"

    tree_value = TreeValue("a")
    tree_value = tree_value.append(TreeValue.empty())
    assert str(tree_value) == "a"
    assert bytes(tree_value) == b"a"
    with pytest.raises(ValueError):
        assert int(tree_value) == 1


def test_tree_value_cross_conversion_from_int():
    # simple case
    tree_value = TreeValue(1)
    assert tree_value.trailing_bits == (1,)
    assert int(tree_value) == 1

    with pytest.raises(ValueError):
        str(tree_value)
    with pytest.raises(ValueError):
        bytes(tree_value)

    # reducing
    tree_value = TreeValue(b"")
    for bit in ONE_BITS:
        tree_value = tree_value.append(TreeValue(int(bit)))  # push 8 bits, or one byte
    assert int(tree_value) == 1
    assert str(tree_value) == "1"
    assert bytes(tree_value) == b"1"


def test_tree_value_cross_conversion_from_str():
    tree_value = TreeValue("Hello, World!")
    with pytest.raises(ValueError):
        int(tree_value)
    assert str(tree_value) == "Hello, World!"
    assert bytes(tree_value) == b"Hello, World!"


def test_tree_value_cross_conversion_from_bytes():
    tree_value = TreeValue(b"Hello, World!")
    with pytest.raises(FandangoConversionError):
        int(tree_value)
    assert str(tree_value) == "Hello, World!"
    assert bytes(tree_value) == b"Hello, World!"

    tree_value = TreeValue(b"123")
    assert int(tree_value) == 123
    assert str(tree_value) == "123"
    assert bytes(tree_value) == b"123"


def test_tree_value_trailing_bits():
    # from string
    tree_value = TreeValue("Hello, World!")
    for bit in A_BITS:
        tree_value = tree_value.append(TreeValue(bit))

    assert str(tree_value) == "Hello, World!a"

    tree_value = TreeValue("Hello, World!")
    for bit in A_BITS:
        tree_value = tree_value.append(TreeValue(bit))
    assert bytes(tree_value) == b"Hello, World!a"

    # from bytes
    tree_value = TreeValue(b"Hello, World!")
    for bit in A_BITS:
        tree_value = tree_value.append(TreeValue(bit))
    assert str(tree_value) == "Hello, World!a"

    tree_value = TreeValue(b"Hello, World!")
    for bit in A_BITS:
        tree_value = tree_value.append(TreeValue(bit))
    assert bytes(tree_value) == b"Hello, World!a"


def test_tree_value_incomplete_trailing_bits():
    tree_value = TreeValue("Hello, World!")
    tree_value = tree_value.append(TreeValue(1))
    with pytest.raises(FandangoConversionError):
        int(tree_value)
    with pytest.raises(FandangoConversionError):
        str(tree_value)
    with pytest.raises(FandangoConversionError):
        bytes(tree_value)


def test_tree_value_combine_with_str():
    # from string
    tree_value = TreeValue("Hello, World!")
    tree_value = tree_value.append(TreeValue("a"))
    assert str(tree_value) == "Hello, World!a"
    assert tree_value.is_type(TreeValueType.STRING)

    # from bytes
    tree_value = TreeValue(b"Hello, World!")
    tree_value = tree_value.append(TreeValue("a"))
    assert bytes(tree_value) == b"Hello, World!a"
    assert tree_value.is_type(TreeValueType.BYTES)

    # from int
    tree_value = TreeValue(1)
    with pytest.raises(FandangoConversionError):
        tree_value = tree_value.append(TreeValue("a"))


def test_tree_value_combine_with_bytes():
    # from bytes
    tree_value = TreeValue(b"Hello, World!")
    tree_value = tree_value.append(TreeValue(b"a"))
    assert bytes(tree_value) == b"Hello, World!a"
    assert tree_value.is_type(TreeValueType.BYTES)

    # from string
    tree_value = TreeValue("Hello, World!")
    tree_value = tree_value.append(TreeValue(b"a"))
    assert bytes(tree_value) == b"Hello, World!a"
    assert tree_value.is_type(TreeValueType.BYTES)

    # from int
    tree_value = TreeValue(1)
    with pytest.raises(FandangoConversionError):
        tree_value = tree_value.append(TreeValue(b"a"))


def test_tree_value_combine_with_int():
    # from string
    tree_value = TreeValue("Hello, World!")
    tree_value = tree_value.append(TreeValue(A_BITS[0]))
    with pytest.raises(ValueError):
        str(tree_value)

    for i in range(1, len(A_BITS)):  # add more bits to be reducible to a byte
        tree_value = tree_value.append(TreeValue(A_BITS[i]))
    assert str(tree_value) == "Hello, World!a"
    for bit in A_BITS:  # add more bits to be reducible to a byte
        tree_value = tree_value.append(TreeValue(bit))
    assert bytes(tree_value) == b"Hello, World!aa"
    assert tree_value.is_type(TreeValueType.BYTES)

    # from bytes
    tree_value = TreeValue(b"Hello, World!")
    tree_value = tree_value.append(TreeValue(A_BITS[0]))
    with pytest.raises(ValueError):
        bytes(tree_value)

    for i in range(1, len(A_BITS)):  # add more bits to be reducible to a byte
        tree_value = tree_value.append(TreeValue(A_BITS[i]))
    assert str(tree_value) == "Hello, World!a"
    for bit in A_BITS:
        tree_value = tree_value.append(TreeValue(bit))
    assert bytes(tree_value) == b"Hello, World!aa"
    assert tree_value.is_type(TreeValueType.BYTES)

    # from int
    tree_value = TreeValue(1)
    tree_value = tree_value.append(TreeValue(1))
    assert int(tree_value) == 3
    assert tree_value.is_type(TreeValueType.TRAILING_BITS_ONLY)


def test_tree_value_to_bits():
    tree_value_str = TreeValue("a")
    assert tree_value_str.to_bits() == f"{ord('a'):08b}"

    tree_value_bytes = TreeValue(b"a")
    assert tree_value_bytes.to_bits() == f"{ord('a'):08b}"

    tree_value_int = TreeValue(1)
    assert tree_value_int.to_bits() == "1"

    tree_value_str = tree_value_str.append(TreeValue(1))
    assert tree_value_str.to_bits() == f"{ord('a'):08b}1"

    tree_value_bytes = tree_value_bytes.append(TreeValue(1))
    assert tree_value_bytes.to_bits() == f"{ord('a'):08b}1"

    tree_value_int = tree_value_int.append(TreeValue(1))
    assert tree_value_int.to_bits() == "11"


def test_direct_access_simple():
    tree_value = TreeValue(1)
    assert -1 == -tree_value  # type: ignore[operator]
    tree = DerivationTree(Terminal(1))
    assert -1 == -tree  # type: ignore[operator]


UNDERLYING_TYPE_NO_ARGS = [
    "__abs__",
    "__ceil__",
    "__float__",
    "__floor__",
    "__index__",
    "__invert__",
    "__neg__",
    "__pos__",
    "__round__",
    "__trunc__",
    "as_integer_ratio",
    "bit_count",
    "bit_length",
    "capitalize",
    "casefold",
    "conjugate",
    "expandtabs",
    "hex",
    "isalnum",
    "isalpha",
    "isascii",
    "isdecimal",
    "isdigit",
    "isidentifier",
    "islower",
    "isnumeric",
    "isprintable",
    "isspace",
    "istitle",
    "isupper",
    "lower",
    "lstrip",
    "split",
    "strip",
    "rsplit",
    "rstrip",
    "splitlines",
    "swapcase",
    "title",
    "upper",
]


@pytest.mark.parametrize("method", UNDERLYING_TYPE_NO_ARGS)
def test_to_underlying_type_no_arg(method):
    def check_method(obj, method):
        if 1 < ((method in dir(1)) + (method in dir("1")) + (method in dir(b"1"))):
            with pytest.warns(DeprecationWarning):
                assert not callable(getattr(obj, method)())
        else:
            assert not callable(getattr(obj, method)())

    run = 0
    if method in dir(1):
        assert not callable(getattr(1, method)())
        check_method(DerivationTree(Terminal(1)), method)
        check_method(TreeValue(1), method)
        run += 1
    if method in dir("1"):
        assert not callable(getattr("1", method)())
        check_method(DerivationTree(Terminal("1")), method)
        check_method(TreeValue("1"), method)
        run += 1
    if method in dir(b"1"):
        assert not callable(getattr(b"1", method)())
        check_method(DerivationTree(Terminal(b"1")), method)
        check_method(TreeValue(b"1"), method)
        run += 1
    assert (
        run > 0
    ), f"{method} not found in dirs of 1, {set(dir(1) + dir('1') + dir(b'1'))}"


UNDERLYING_TYPE_INT_ARG = [
    "__mul__",
    "__pow__",
    "__rmul__",
    "__rpow__",
    "center",
    "ljust",
    "rjust",
    "zfill",
]


@pytest.mark.parametrize("method", UNDERLYING_TYPE_INT_ARG)
def test_to_underlying_type_int_arg(method):
    left: tuple[int | str | bytes, ...] = (1, "1", b"1")
    right: tuple[int | str | bytes, ...] = (1, 1, 1)
    if "__r" in method:
        left, right = right, left

    def check_method(obj, method, *args):
        if 1 < ((method in dir(1)) + (method in dir("1")) + (method in dir(b"1"))):
            with pytest.warns(DeprecationWarning):
                assert not callable(getattr(obj, method)(*args))
        else:
            assert not callable(getattr(obj, method)(*args))

    run = 0
    if method in dir(1):
        assert not callable(getattr(left[0], method)(right[0]))
        check_method(DerivationTree(Terminal(left[0])), method, right[0])
        check_method(TreeValue(left[0]), method, right[0])
        run += 1
    if method in dir("1"):
        assert not callable(getattr(left[1], method)(right[1]))
        check_method(DerivationTree(Terminal(left[1])), method, right[1])
        check_method(TreeValue(left[1]), method, right[1])
        run += 1
    if method in dir(b"1"):
        assert not callable(getattr(left[2], method)(right[2]))
        check_method(DerivationTree(Terminal(left[2])), method, right[2])
        check_method(TreeValue(left[2]), method, right[2])
        run += 1
    assert (
        run > 0
    ), f"{method} not found in dirs of 1, {set(dir(1) + dir('1') + dir(b'1'))}"


MODS = [
    "__mod__",
    "__rmod__",
]


@pytest.mark.parametrize("method", MODS)
def test_to_underlying_type_mod(method):
    run = 0
    left = (1, "%s", b"%s")
    right = (1, "1", b"1")
    if "r" in method:
        left, right = right, left

    if method in dir(1):
        assert not callable(getattr(left[0], method)(right[0]))
        assert not callable(
            getattr(DerivationTree(Terminal(left[0])), method)(right[0])
        )
        assert not callable(getattr(TreeValue(left[0]), method)(right[0]))
        run += 1
    if method in dir("1"):
        assert not callable(getattr(left[1], method)(right[1]))
        assert not callable(
            getattr(DerivationTree(Terminal(left[1])), method)(right[1])
        )
        assert not callable(getattr(TreeValue(left[1]), method)(right[1]))
        run += 1
    if method in dir(b"1"):
        assert not callable(getattr(left[2], method)(right[2]))
        assert not callable(
            getattr(DerivationTree(Terminal(left[2])), method)(right[2])
        )
        assert not callable(getattr(TreeValue(left[2]), method)(right[2]))
        run += 1
    assert (
        run > 0
    ), f"{method} not found in dirs of 1, {set(dir(1) + dir('1') + dir(b'1'))}"


FORMATS = [
    "format",
]


@pytest.mark.parametrize("method", FORMATS)
def test_to_underlying_type_format(method):
    def check_method(obj, method, *args, **kwargs):
        if 1 < ((method in dir(1)) + (method in dir("1")) + (method in dir(b"1"))):
            with pytest.warns(DeprecationWarning):
                assert not callable(getattr(obj, method)(*args, **kwargs))
        else:
            assert not callable(getattr(obj, method)(*args, **kwargs))

    run = 0
    if method in dir(1):
        assert not callable(getattr(1, method)())
        check_method(DerivationTree(Terminal(1)), method)
        check_method(TreeValue(1), method)
        run += 1
    if method in dir("1"):
        assert not callable(getattr("{}", method)("1"))
        check_method(DerivationTree(Terminal("{}")), method, "1")
        check_method(TreeValue("{}"), method, "1")
        run += 1
    if method in dir(b"1"):
        assert not callable(getattr(b"{}", method)(b"1"))
        check_method(DerivationTree(Terminal(b"{}")), method, b"1")
        check_method(TreeValue(b"{}"), method, b"1")
        run += 1

    assert (
        run > 0
    ), f"{method} not found in dirs of 1, {set(dir(1) + dir('1') + dir(b'1'))}"


FORMAT_MAP = [
    "format_map",
]


@pytest.mark.parametrize("method", FORMAT_MAP)
def test_to_underlying_type_format_map(method):
    def check_method(obj, method, *args, **kwargs):
        if 1 < ((method in dir(1)) + (method in dir("1")) + (method in dir(b"1"))):
            with pytest.warns(DeprecationWarning):
                assert not callable(getattr(obj, method)(*args, **kwargs))
        else:
            assert not callable(getattr(obj, method)(*args, **kwargs))

    run = 0
    if method in dir(1):
        assert not callable(getattr(1, method)({1: 1}))
        check_method(DerivationTree(Terminal(1)), method, {1: 1})
        check_method(TreeValue(1), method, {1: 1})
        run += 1
    if method in dir("1"):
        assert not callable(getattr("1", method)({"1": "1"}))
        check_method(DerivationTree(Terminal("1")), method, {"1": "1"})
        check_method(TreeValue("1"), method, {"1": "1"})
        run += 1
    if method in dir(b"1"):
        assert not callable(getattr(b"1", method)({b"1": b"1"}))
        check_method(DerivationTree(Terminal(b"1")), method, {b"1": b"1"})
        check_method(TreeValue(b"1"), method, {b"1": b"1"})


FIRST_ARG = [
    "__add__",
    "__and__",
    "__contains__",
    "__divmod__",
    "__floordiv__",
    "__ge__",
    "__gt__",
    "__le__",
    "__lshift__",
    "__lt__",
    "__or__",
    "__radd__",
    "__rand__",
    "__rdivmod__",
    "__rfloordiv__",
    "__rlshift__",
    "__ror__",
    "__rrshift__",
    "__rshift__",
    "__rsub__",
    "__rtruediv__",
    "__rxor__",
    "__sub__",
    "__truediv__",
    "__xor__",
    "count",
    "endswith",
    "find",
    "index",
    "partition",
    "removeprefix",
    "removesuffix",
    "rfind",
    "rindex",
    "rpartition",
    "startswith",
]


@pytest.mark.parametrize("method", FIRST_ARG)
def test_to_first_arg(method):
    run = 0
    if method in dir(1):
        assert not callable(getattr(1, method)(1))
        assert not callable(getattr(DerivationTree(Terminal(1)), method)(1))
        assert not callable(getattr(TreeValue(1), method)(1))
        run += 1
    if method in dir("1"):
        assert not callable(getattr("1", method)("1"))
        assert not callable(getattr(DerivationTree(Terminal("1")), method)("1"))
        assert not callable(getattr(TreeValue("1"), method)("1"))
        run += 1
    if method in dir(b"1"):
        assert not callable(getattr(b"1", method)(b"1"))
        assert not callable(getattr(DerivationTree(Terminal(b"1")), method)(b"1"))
        assert not callable(getattr(TreeValue(b"1"), method)(b"1"))
        run += 1
    assert (
        run > 0
    ), f"{method} not found in dirs of 1, {set(dir(1) + dir('1') + dir(b'1'))}"


def test_check_all_direct_methods_tested():
    in_tests = (
        FIRST_ARG
        + UNDERLYING_TYPE_NO_ARGS
        + UNDERLYING_TYPE_INT_ARG
        + MODS
        + FORMATS
        + FORMAT_MAP
    )
    implemented = (
        DIRECT_ACCESS_METHODS_BASE_TO_FIRST_ARG_TYPE
        + DIRECT_ACCESS_METHODS_BASE_TO_UNDERLYING_TYPE
    )
    assert sorted(in_tests) == sorted(implemented)


@pytest.mark.parametrize(
    "base_value",
    [TreeValue("Hello, World!"), DerivationTree(Terminal("Hello, World!"))],
)
def test_tree_value_direct_access_non_base_type(base_value):
    assert base_value.startswith("Hello")
    with pytest.warns(DeprecationWarning):
        assert base_value.startswith(TreeValue("Hello"))
    with pytest.warns(DeprecationWarning):
        assert base_value.startswith(DerivationTree(Terminal("Hello")))

    with pytest.raises(Exception):
        base_value.startswith(1.0)  # float is illegal base type



from fandango.language.grammar.grammar import Grammar
from fandango.language.grammar.nodes.node import NodeType
from fandango.language.search import NonTerminalSearch
from fandango.language.symbols import NonTerminal, Terminal, Symbol, SymbolType, Slice
from fandango.language.tree import DerivationTree

__all__ = [
    "Symbol",
    "Terminal",
    "NonTerminal",
    "Slice",
    "SymbolType",
    "DerivationTree",
    "Grammar",
    "NonTerminalSearch",
    "NodeType",
]



# wildcard import required for usage in spec files
from fandango.io import *  # noqa: F403
from fandango.language.symbols import NonTerminal
from fandango.language.tree import DerivationTree


def is_before(
    tree: DerivationTree, before_tree: DerivationTree, after_tree: DerivationTree
):
    """
    Check if the tree is before the before_tree and after the after_tree.
    """
    before_index = tree.get_index(before_tree)
    after_index = tree.get_index(after_tree)
    if before_index < 0 or after_index < 0:
        return False
    return before_index < after_index


def is_after(
    tree: DerivationTree, after_tree: DerivationTree, before_tree: DerivationTree
):
    """
    Check if the tree is after the after_tree and before the before_tree.
    """
    return is_before(tree, before_tree, after_tree)


def get_index_within(
    tree: DerivationTree, scope: DerivationTree, index_counter_symbols: list[str]
) -> int:
    idx = 0
    index_counter_nts = [NonTerminal(symbol) for symbol in index_counter_symbols]
    for val in scope.flatten():
        if val == tree:
            return idx
        if val.symbol in index_counter_nts:
            idx += 1
    return -1



import logging
import re
from typing import Counter

import pytest
from fandango.language.grammar.grammar import Grammar
from fandango.language.grammar.grammar_settings import GrammarSetting
from fandango.language.grammar.nodes.alternative import Alternative
from fandango.language.grammar.nodes.terminal import TerminalNode
from fandango.language.parse import parse
from fandango.language.symbols.non_terminal import NonTerminal
from fandango.language.symbols.terminal import Terminal
from fandango.language.tree import DerivationTree

REPETITIONS = 1000


# Gmutator mutation (1a)
def test_repeat_terminal():
    raw_grammar = """
<start> ::= "a"
setting all_with_type(TerminalNode) terminal_should_repeat = 1.0
"""
    grammar, _ = parse(raw_grammar)
    assert grammar is not None
    found_empty = False
    found_multiple = False
    for _ in range(REPETITIONS):
        tree = grammar.fuzz()
        assert tree is not None
        tree_str = str(tree)
        num_as = tree_str.count("a")
        assert len(tree_str) == num_as
        if num_as > 1:
            found_multiple = True
        elif num_as == 0:
            found_empty = True
        else:
            assert False, f"Unexpected tree: {tree_str}"
    assert found_empty
    assert found_multiple


# Gmutator mutation (1b)
def test_plus_should_return_nothing():
    raw_grammar = """
<start> ::= <nonterminal>+
<nonterminal> ::= "a"
setting all_with_type(Plus) plus_should_return_nothing = 1.0
"""
    grammar, _ = parse(raw_grammar)
    assert grammar is not None

    for _ in range(REPETITIONS):
        tree = grammar.fuzz()
        assert tree is not None
        assert tree.to_string() == ""


# Gmutator mutation (1c)
def test_option_should_return_multiple():
    raw_grammar = """
<start> ::= <nonterminal>?
<nonterminal> ::= "a"
setting all_with_type(Option) option_should_return_multiple = 1.0
"""
    grammar, _ = parse(raw_grammar)
    assert grammar is not None
    for _ in range(REPETITIONS):
        tree = grammar.fuzz()
        assert tree is not None
        t = tree.to_string()
        assert len(t) >= 2
        assert t.count("a") == len(t)


# Gmutator mutation (2)
def test_concatenated_alternatives():
    raw_grammar = """
<start> ::= <a> | <b>
<a> ::= "a"
<b> ::= "b"
setting all_with_type(Alternative) alternatives_should_concatenate = 1.0
"""
    grammar, _ = parse(raw_grammar)
    assert grammar is not None

    for _ in range(REPETITIONS):
        tree = grammar.fuzz()
        assert tree is not None
        if tree.to_string() == "ab":
            found_ab = True
        elif tree.to_string() == "ba":
            found_ba = True
        else:
            raise ValueError(f"Unexpected tree: {tree.to_string()}")
    assert found_ab
    assert found_ba


# Gmutator mutation (2)
def test_concatenated_alternatives_with_multiple_alternatives():
    raw_grammar = """
<start> ::= <a> | <b> | <c> | <d>
<a> ::= "a"
<b> ::= "b"
<c> ::= "c"
<d> ::= "d"
setting all_with_type(Alternative) alternatives_should_concatenate = 1.0
"""
    grammar, _ = parse(raw_grammar)
    assert grammar is not None

    for _ in range(REPETITIONS):
        tree = grammar.fuzz()
        s = tree.to_string()
        assert len(s) >= 2
        assert all(c in "abcd" for c in s)
        if Counter(s) == {"a": 1, "b": 1, "c": 1, "d": 1}:
            found_abcd = True
    assert found_abcd


# Gmutator mutation (2)
def test_concatenated_alternatives_with_one_node():
    outer = DerivationTree(NonTerminal("<start>"))
    settings = GrammarSetting(
        "all_with_type(Alternative)", {"alternatives_should_concatenate": "1.0"}
    )
    alt = Alternative([TerminalNode(Terminal("a"), [])], [settings])
    alt.distance_to_completion = 2
    alt.fuzz(outer, Grammar.dummy(), 100)
    assert outer.to_string() == "a"


# Gmutator mutation (3)
def test_invert_regex():
    pattern = r"a|b|c"
    raw_grammar = f"""
<start> ::= r"{pattern}"
setting all_with_type(TerminalNode) invert_regex = 1.0
"""
    grammar, _ = parse(raw_grammar)
    assert grammar is not None

    for _ in range(REPETITIONS):
        tree = grammar.fuzz()
        assert tree is not None
        s = tree.to_string()
        assert re.match(pattern, s) is None  # should *not* match


# Gmutator mutation (3)
def test_max_out_of_regex_tries(caplog: pytest.LogCaptureFixture):
    pattern = r"a|b|c"
    raw_grammar = f"""
<start> ::= r"{pattern}"
setting all_with_type(TerminalNode) invert_regex = 1.0
setting all_with_type(TerminalNode) max_out_of_regex_tries = 0
"""
    grammar, _ = parse(raw_grammar)
    assert grammar is not None

    for _ in range(REPETITIONS):
        caplog.set_level(logging.WARNING)
        caplog.clear()
        tree = grammar.fuzz()
        assert tree is not None
        out_list = caplog.get_records("call")

        assert len(out_list) == 1
        message = out_list[0].message
        assert (
            f"Failed to generate a non-matching regex: {pattern}, falling back to matching regex"
            in message
        ), message

        s = tree.to_string()
        assert (
            re.match(pattern, s) is not None
        )  # should match because we're out of tries, so we're falling back to matching regex


def test_impossible_out_of_regex_pattern(caplog: pytest.LogCaptureFixture):
    pattern = r".*"
    raw_grammar = f"""
<start> ::= r"{pattern}"
setting all_with_type(TerminalNode) invert_regex = 1.0
setting all_with_type(TerminalNode) max_out_of_regex_tries = 100
"""
    grammar, _ = parse(raw_grammar)
    assert grammar is not None

    for _ in range(REPETITIONS):
        caplog.set_level(logging.WARNING)
        caplog.clear()
        tree = grammar.fuzz()
        assert tree is not None
        out_list = caplog.get_records("call")

        assert len(out_list) == 1
        message = out_list[0].message
        assert (
            f"Failed to generate a non-matching regex: {pattern}, falling back to matching regex"
            in message
        ), message

        s = tree.to_string()
        assert (
            re.match(pattern, s) is not None
        )  # should match because we're out of tries, so we're falling back to matching regex


# Gmutator mutation (4)
def test_non_terminal_use_other_rule():
    raw_grammar = """
<start> ::= "prefix" <a>
<a> ::= "a"
setting <start> non_terminal_use_other_rule = 1.0
"""
    grammar, _ = parse(raw_grammar, use_stdlib=False)
    assert grammar is not None

    for _ in range(1):
        tree = grammar.fuzz()
        assert tree is not None
        s = tree.to_string()
        assert s == "a"



#!/usr/bin/env pytest
import asyncio
import os
import re
import shutil
import subprocess
import sys
import time
import unittest

from fandango.cli import get_parser

from .utils import DOCS_ROOT, IS_BEARTYPE_ACTIVE, RESOURCES_ROOT, run_command

# beartype somehow scrambles the fixed rng
if IS_BEARTYPE_ACTIVE:
    expected_with_random_seed = [
        "6040",
        "0561",
        "37900",
        "17",
        "06",
        "9",
        "56683",
        "8254",
        "44066",
        "52",
    ]
else:
    expected_with_random_seed = [
        "35716",
        "4",
        "9768",
        "30",
        "5658",
        "5",
        "9",
        "649",
        "20",
        "41",
    ]


class TestCLI(unittest.TestCase):
    def test_help(self):
        command = ["fandango", "--help"]
        out, err, code = run_command(command)
        _parser = get_parser(True)
        self.assertEqual(0, code)
        self.assertEqual(err, "")

    def test_fuzz_basic(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(RESOURCES_ROOT / "digit.fan"),
            "-n",
            "10",
            "--random-seed",
            "426912",
            "--no-cache",
        ]
        out, err, code = run_command(command)
        self.assertEqual(0, code)
        self.assertEqual(err, "")
        self.assertEqual(expected_with_random_seed, out.strip().split("\n"))

    def test_output_to_file(self):
        out_file = RESOURCES_ROOT / "test.txt"
        if os.path.exists(out_file):
            os.remove(out_file)
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(RESOURCES_ROOT / "digit.fan"),
            "-n",
            "10",
            "--random-seed",
            "426912",
            "-o",
            str(out_file),
            "-s",
            ";",
            "--no-cache",
        ]
        out, err, code = run_command(command)
        self.maxDiff = 1000000
        self.assertEqual(0, code)
        self.assertEqual("", out)
        self.assertEqual("", err)
        with open(out_file, "r") as fd:
            actual = fd.read()
        self.assertEqual(expected_with_random_seed, actual.split(";"))
        os.remove(RESOURCES_ROOT / "test.txt")

    def test_output_multiple_files(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(RESOURCES_ROOT / "digit.fan"),
            "-n",
            "10",
            "--random-seed",
            "426912",
            "-d",
            str(RESOURCES_ROOT / "test"),
            "--no-cache",
        ]
        (
            out,
            err,
            code,
        ) = run_command(command)
        self.assertEqual(0, code)
        self.assertEqual("", out)
        self.assertEqual("", err)
        for i, expected_value in enumerate(expected_with_random_seed):
            filename = RESOURCES_ROOT / "test" / f"fandango-{i:04d}.txt"
            with open(filename, "r") as fd:
                actual = fd.read()
            self.assertEqual(expected_value, actual)

        shutil.rmtree(RESOURCES_ROOT / "test", ignore_errors=True)

    def test_output_with_libfuzzer_harness(self):
        if sys.platform.startswith("win"):
            self.skipTest("LibFuzzer interface not supported on Windows")
        output_file = str(RESOURCES_ROOT / "test_libfuzzer_interface")
        compile_ = [
            "clang",
            "-g",
            "-O2",
            "-fPIC",
            "-shared",
            "-o",
            output_file,
            str(RESOURCES_ROOT / "test_libfuzzer_interface.c"),
        ]
        out, err, code = run_command(compile_)

        self.assertEqual("", out)
        self.assertEqual("", err)
        self.assertEqual(0, code)

        command = [
            "fandango",
            "fuzz",
            "-f",
            str(RESOURCES_ROOT / "digit.fan"),
            "-n",
            "10",
            "--random-seed",
            "426912",
            "--file-mode",
            "binary",
            "--no-cache",
            "--input-method",
            "libfuzzer",
            output_file,
        ]
        expected_output = (
            "\n".join([f"data: {value}" for value in expected_with_random_seed]) + "\n"
        )
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual(expected_output, out)
        self.assertEqual(0, code)

    def test_infinite_mode(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(RESOURCES_ROOT / "digit.fan"),
            "--infinite",
            "--no-cache",
        ]
        proc = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            bufsize=0,  # Unbuffered
            env={**os.environ, "PYTHONUNBUFFERED": "1"},  # Force Python unbuffered
        )
        time.sleep(20)
        self.assertIsNone(proc.poll(), "Process terminated before 20 seconds")
        proc.terminate()
        out, _ = proc.communicate()
        printed_lines = out.splitlines()
        self.assertGreater(
            len(printed_lines),
            100,
            f"Not enough output lines: {len(printed_lines)}",
        )

    def test_unsat(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(RESOURCES_ROOT / "digit.fan"),
            "-n",
            "10",
            "--random-seed",
            "426912",
            "-c",
            "False",
            "--max-generations",
            "50",
        ]
        expected = """fandango:ERROR: Population did not converge to a perfect population
fandango:ERROR: Only found 0 perfect solutions, instead of the required 10
"""
        out, err, code = run_command(command)
        self.assertEqual(0, code)
        self.assertEqual("", out)
        self.assertEqual(expected, err)

    def test_binfinity(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(DOCS_ROOT / "binfinity.fan"),
            "-n",
            "1",
            "--format=none",
            "--validate",
            "--random-seed",
            "426912",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_infinity(self):
        # docs/infinity.fan can only generate a limited number of individuals,
        # so we decrease the population size
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(DOCS_ROOT / "infinity.fan"),
            "-n",
            "1",
            "--format=none",
            "--validate",
            "--random-seed",
            "426912",
            "--population-size",
            "10",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_max_repetition(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(RESOURCES_ROOT / "digit.fan"),
            "-n",
            "10",
            "--max-generations",
            "50",
            "--max-repetitions",
            "10",
            "--no-cache",
            "-c",
            "len(str(<start>)) > 10",
        ]
        expected = """fandango:ERROR: Population did not converge to a perfect population
fandango:ERROR: Only found 0 perfect solutions, instead of the required 10
"""
        out, err, code = run_command(command)
        self.assertEqual(0, code)
        self.assertEqual("", out)
        self.assertEqual(expected, err)

    def test_max_nodes_unsat(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(RESOURCES_ROOT / "gen_number.fan"),
            "-n",
            "10",
            "--population-size",
            "10",
            "--max-generations",
            "30",
            "--no-cache",
            "-c",
            "len(str(<start>)) > 60",
            "--max-nodes",
            "30",
        ]
        err_pattern = r"""fandango:ERROR: Population did not converge to a perfect population
fandango:ERROR: Only found (\d) perfect solutions, instead of the required 10"""
        out_pattern = (
            r"""(aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa\.\d+\n)*"""
        )
        out, err, code = run_command(command)
        self.assertRegex(out, out_pattern, f"out: {out}")
        self.assertRegex(err, err_pattern, f"err: {err}")
        self.assertEqual(0, code)

        num_from_error_message = int(re.findall(err_pattern, err)[0])
        self.assertEqual(num_from_error_message, len(out.split("\n")) - 1)

    def test_unparse_grammar(self):
        # We unparse the standard library as well as docs/persons.fan
        input_data = f"set -f {DOCS_ROOT / 'persons.fan'}\nset\n"
        out, err, code = run_command(["fandango", "shell"], input=input_data)
        self.assertEqual(0, code)
        self.maxDiff = 1000000
        self.assertEqual("", err)
        self.assertTrue(out.startswith("<_char> ::= r'(.|\\n)'\n"), out)
        self.assertTrue(out.endswith("<age> ::= <digit>+\n"), out)

    def test_talk_cat(self):
        command = [
            "fandango",
            "-v",
            "talk",
            "-n",
            "1",
            "-f",
            str(DOCS_ROOT / "cat-oracle.fan"),
            "cat",
        ]
        out, err, code = run_command(command)
        split_err = err.split("\n")

        filter_prefixes = ["fandango:INFO: In:", "fandango:INFO: Out:"]
        io_logs = list(
            filter(
                lambda x: any(filter(lambda b: x.startswith(b), filter_prefixes)),
                split_err,
            )
        )
        self.assertEqual(2, len(io_logs), f"err: {err}")
        result_a = io_logs[0].split(": ", 2)[2]
        result_b = io_logs[0].split(": ", 2)[2]
        self.assertEqual(result_a, result_b)
        self.assertEqual(0, code)
        self.assertEqual("", out)

    def test_soliloquy(self):
        async def async_run():
            def run_server():
                server_cmd = [
                    "fandango",
                    "-v",
                    "talk",
                    "-n",
                    "1",
                    "-f",
                    str(DOCS_ROOT / "smtp-extended.fan"),
                    "--server",
                    "tcp://localhost:9025",
                ]
                out, err, code = run_command(server_cmd)
                return out, err, code

            def run_client():
                import time

                time.sleep(
                    20
                )  # delay to let server start. We should find a better method for this
                client_cmd = [
                    "fandango",
                    "-v",
                    "talk",
                    "-n",
                    "1",
                    "-f",
                    str(DOCS_ROOT / "smtp-extended.fan"),
                    "--client",
                    "tcp://localhost:9025",
                ]
                out, err, code = run_command(client_cmd)
                return out, err, code

            # Run both in threads (since self.run_command is sync)
            server_future = asyncio.to_thread(run_server)
            client_future = asyncio.to_thread(run_client)
            return await asyncio.gather(server_future, client_future)

        (server_out, server_err, server_code), (client_out, client_err, client_code) = (
            asyncio.run(async_run())
        )
        self.assertEqual(
            0,
            server_code,
            f"Server error: {server_err}\n\nServer output: {server_out}",
        )
        self.assertEqual(
            0,
            client_code,
            f"Client error: {client_err}\n\nClient output: {client_out}",
        )



from typing import TYPE_CHECKING

from fandango.constraints.constraint import Constraint

if TYPE_CHECKING:
    from fandango.constraints.comparison import ComparisonConstraint
    from fandango.constraints.conjunction import ConjunctionConstraint
    from fandango.constraints.disjunct import DisjunctionConstraint
    from fandango.constraints.expression import ExpressionConstraint
    from fandango.constraints.forall import ForallConstraint
    from fandango.constraints.exists import ExistsConstraint
    from fandango.constraints.implication import ImplicationConstraint
    from fandango.constraints.repetition_bounds import (
        RepetitionBoundsConstraint,
    )


class ConstraintVisitor:
    """
    A base class for visiting and processing different types of constraints.

    This class uses the visitor pattern to traverse constraint structures. Each method
    corresponds to a specific type of constraint, allowing implementations to define
    custom behavior for processing or interacting with that type.
    """

    def __init__(self):
        pass

    def do_continue(self, constraint: "Constraint") -> bool:
        """If this returns False, this formula should not call the visit methods for
        its children."""
        return True

    def visit(self, constraint: "Constraint"):
        """Visits a constraint."""
        return constraint.accept(self)

    def visit_expression_constraint(self, constraint: "ExpressionConstraint"):
        """Visits an expression constraint."""
        pass

    def visit_comparison_constraint(self, constraint: "ComparisonConstraint"):
        """Visits a comparison constraint."""
        pass

    def visit_forall_constraint(self, constraint: "ForallConstraint"):
        """Visits a forall constraint."""
        pass

    def visit_exists_constraint(self, constraint: "ExistsConstraint"):
        """Visits an exists constraint."""
        pass

    def visit_disjunction_constraint(self, constraint: "DisjunctionConstraint"):
        """Visits a disjunction constraint."""
        pass

    def visit_conjunction_constraint(self, constraint: "ConjunctionConstraint"):
        """Visits a conjunction constraint."""
        pass

    def visit_implication_constraint(self, constraint: "ImplicationConstraint"):
        """Visits an implication constraint."""
        pass

    def visit_repetition_bounds_constraint(
        self, constraint: "RepetitionBoundsConstraint"
    ):
        """Visits a repetition bounds constraint."""
        pass



def activate_beartype() -> None:
    from beartype.claw import beartype_this_package  # type: ignore [import-not-found]
    from beartype import BeartypeConf  # type: ignore [import-not-found]

    skip_packages = (
        "fandango.converters.antlr.ANTLRv4Parser",  # auto-generated
        "fandango.converters.antlr.ANTLRv4Lexer",  # auto-generated
        "fandango.language.grammar.parser.iterative_parser",
        "fandango.language.parser",  # broken
        "fandango.language.search",  # broken, at least test_item_search and test_searches call ItemSearch with non-lists
        "fandango.constraints.fitness",  # ValueFitness sometimes receives a list of ints in the constructor
        "fandango.io.packetforecaster",  # broken
    )

    beartype_this_package(conf=BeartypeConf(claw_skip_package_names=skip_packages))
